package CMapParsing;

#
#   Author:     Gurmukh Sahota 07/15/2000
#   Copyright:  Gurmukh Sahota, 2000. All rights reserved.
#
#   Modified : Gurmukh Sahota 06/17/2001
#              messed with reMultiply so that incase a null cmap_hlist is passed in, the sub
#              does not fail, but will create nulls for the data.  Then allowed export of subroutine.
#   Copyright:  Gurmukh Sahota 2001.  All rights reserved.
#
#

# Returns a reference to a Array     of  a Hash      of an Array
#                          Residue  :::  AtomTypes  :::  possible chemical shifts

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw( clone parseCMapfile outputCMapfile reMultiply deMultiplicate readBMRBasCMap convertBMRBtoCMap);
%EXPORT_TAGS = ( ALL => [@EXPORT_OK] );                                                   

# allows perl to find local modules.
use FindBin;
use lib $FindBin::Bin;
use BMRBParsing qw(:ALL);

use strict;


my %multiplicity = ("A" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "HA" => 1, "HB" => 1},
		    "C" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 1, "SG" => 1},
		    "D" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HD" => 1, "HD2" => 1, "OD" => 2, "OD1" => 1, "OD2" => 1},
		    "E" => { "HN" => 1, "N15" => 1, "N" => 1,  "NE" => 1, "NE2" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 2, "HG2" => 1, "HG3" => 1, "HE" => 2, "HE21" => 1, "HE22" => 1, "OE" => 1, "OE1" => 1},
		    "F" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 2, "CD1" => 1, "CD2" => 1, "CE" => 2, "CE1" => 1, "CE2" => 1, "CZ" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HD" => 2, "HD1" => 1, "HD2" => 1, "HE" => 2, "HE1" => 1, "HE2" => 1, "HZ" => 1},
		    "G" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "HA" => 2, "HA2" => 1, "HA3" => 1 },
		    "H" => { "HN" => 1, "N15" => 1, "N" => 1, "ND" => 1, "ND1" => 1, "NE" => 1, "NE2" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 1, "CD2" => 1, "CE" => 1, "CE1" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HD" => 2, "HD1" => 1, "HD2" => 1, "HE" => 2, "HE1" => 1, "HE2" => 1},
		    "I" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 2, "CG1" => 1, "CG2" => 1, "CD" => 1, "CD1" => 1, "HA" => 1, "HB" => 1, "HB" => 1, "HG" => 3, "HG1" => 2, "HG2" => 1, "HG12" => 1, "HG13" => 1, "HD" => 1, "HD1" => 1},
		    "K" => { "HN" => 1, "N15" => 1, "N" => 1, "NZ" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 1, "CE" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 2, "HG2" => 1, "HG3" => 1, "HD" => 2, "HD2" => 1, "HD3" => 1, "HE" => 2, "HE2" => 1, "HE3" => 1, "HZ" => 1},
		    "L" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 2, "CD1" => 1, "CD2" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 1, "HD" => 2, "HD1" => 1, "HD2" => 1},
		    "M" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CE" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 2, "HG2" => 1, "HG3" => 1, "HE" => 1, "SD" => 1},
		    "N" => { "HN" => 1, "N15" => 1, "N" => 1, "ND" => 1, "ND2" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HD" => 2, "HD2" => 2, "HD21" => 1, "HD22" => 1, "OD" => 1, "OD1" => 1},
		    "P" => {            "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 2, "HG2" => 1, "HG3" => 1, "HD" => 2, "HD2" => 1, "HD3" => 1},
		    "Q" => { "HN" => 1, "N15" => 1, "N" => 1, "NE" => 1, "NE2" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 2, "HG2" => 1, "HG3" => 1 , "HE" => 2, "HE2" => 2, "HE21" => 1, "HE22" => 1, "OE" => 1, "OE1" => 1},
		    "R" => { "HN" => 1, "N15" => 1, "N" => 1, "NE" => 1, "NH" => 2, "NH1" => 1, "NH2" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 1, "CZ" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 1, "HD" => 2, "HD1" => 1, "HD2" => 1, "HE"  => 1, "HH" => 4, "HH1" => 2, "HH2" => 2, "HH11" => 1, "HH12" => 1, "HH21" => 1, "HH22" => 1},
		    "S" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HG" => 1, "OG" => 1},
		    "T" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CG2" => 1, "HA" => 1, "HB" => 1, "HG" => 2, "HG1" => 1, "HG2" => 1, "OG" => 1, "OG1" => 1},
		    "V" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 2, "CG1" => 1, "CG2" => 1, "HA" => 1, "HB" => 1, "HG" => 2, "HG1" => 1, "HG2" => 1},
		    "W" => { "HN" => 1, "N15" => 1, "N" => 1, "NE" => 1, "NE1" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 2, "CD1" => 1, "CD2" => 1, "CE" => 2, "CE2" => 1, "CE3" => 1, "CZ" => 2, "CZ2" => 1, "CZ3" => 1, "CH" => 1, "CH2" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HD" => 1, "HD1" => 1, "HE" => 2, "HE1" => 1, "HE3" => 1, "HZ" => 2, "HZ2" => 1, "HZ3" => 1, "HH" => 1, "HH2" =>1},

		    "Y" => { "HN" => 1, "N15" => 1, "N" => 1, "CO" => 1, "C" => 1, "CA" => 1, "CB" => 1, "CG" => 1, "CD" => 2, "CD1" => 1, "CD2" => 1, "CE" => 2, "CE1" => 1, "CE2" => 1, "CZ" => 1, "HA" => 1, "HB" => 2, "HB2" => 1, "HB3" => 1, "HD" => 2, "HD1" => 1, "HD2" => 2, "HE" => 2, "HE1" => 1, "HE2" => 2, "HZ" => 1, "HH" => 1, "OH" => 1}
		    );


# clone
#   clones a cmap_hlist.
#
#        $original_hlist - the original $bmrb_hlist
#        $cloned_hlist   - the cloned $bmrb_hlist
sub clone
  {
  my $original_hlist = shift @_;
  my $cloned_hlist = shift @_;
  
  # generic cloning code
  my @clone_stack;
  push @clone_stack, { "type" => "HASH", "clone" => $cloned_hlist, "range" => [ keys %$original_hlist ], "index" => 0, "original" => $original_hlist }; 
  
  while(@clone_stack)
    {
    my $index = $clone_stack[$#clone_stack]{"index"};

    if ($index < @{$clone_stack[$#clone_stack]{"range"}}) # clone a part of the current hash or array
      {
      # get new value
      my $test_value;
      if ($clone_stack[$#clone_stack]{"type"} eq "HASH")
	{ $test_value = $clone_stack[$#clone_stack]{"original"}{$clone_stack[$#clone_stack]{"range"}[$index]}; }
      elsif ($clone_stack[$#clone_stack]{"type"} eq "ARRAY")
	{ $test_value = $clone_stack[$#clone_stack]{"original"}[$index]; }

      my $test_ref = ref($test_value);
      if ($test_ref eq "HASH") # test value is a hash
	{ push @clone_stack, {"type" => "HASH", "clone" => {}, "range" => [ keys %$test_value ], "index" => 0, "original" => $test_value }; }
      elsif ($test_ref eq "ARRAY") # test value is an array
	{ push @clone_stack, {"type" => "ARRAY", "clone" => [], "range" => $test_value, "index" => 0, "original" => $test_value }; }
      elsif ($clone_stack[$#clone_stack]{"type"} eq "HASH") # clone section is a hash
	{
	$clone_stack[$#clone_stack]{"clone"}{$clone_stack[$#clone_stack]{"range"}[$index]} = $test_value;
	$clone_stack[$#clone_stack]{"index"}++;
	}
      else # clone section is an array
	{
	$clone_stack[$#clone_stack]{"clone"}[$index] = $test_value;
	$clone_stack[$#clone_stack]{"index"}++;
	}
      }
    else # pop finished cloned section.
      {
      if (@clone_stack > 1)
	{
	my $index = $clone_stack[$#clone_stack - 1]{"index"};

	if($clone_stack[$#clone_stack - 1]{"type"} eq "HASH") # handle hashes
	  { $clone_stack[$#clone_stack - 1]{"clone"}{$clone_stack[$#clone_stack - 1]{"range"}[$index]} = $clone_stack[$#clone_stack]{"clone"}; }
	else # handle arrays
	  { $clone_stack[$#clone_stack - 1]{"clone"}[$index] = $clone_stack[$#clone_stack]{"clone"}; }
	}
      
      pop @clone_stack;
      if (@clone_stack)
	{ $clone_stack[$#clone_stack]{"index"}++; }
      }
    }
  
  
  return;
  }

####################################### parseCMapfile #############################################
# Input   : cmap filename to parse into below described datastructure                             # 
# Output  : A reference to a cmap hash datastructure                                              #
#                $cmap_hlist - reference to a hash                                                #
#               $$cmap_hlist{"filename"}    - cmap's filename                                     #
#               $$cmap_hlist{"name_array"}  - list of residue names in their proper order         #
#               $$cmap_hlist{"parsedtokens"}- the tokens of the cmapfile usually AtomTypes        #
#                                             (HN, N15, CA, CA-1) and also AA for AA#             #
#               $$cmap_hlist{"sections"}    - array of array of parsed tokens in each section     #
#               $$cmap_hlist{"settings"}    - array of settings lines                             #
#               $$cmap_hlist{"comments"}    - the comments in the cmap file, lines                #
#                                             starting with a "#"                                 #
#               $$cmap_hlist{"sequence"}    - sequence of protein describe in cmapfile            #
#               $$cmap_hlist{"sequence_startpos"} - starting residue number of sequence of        #
#                                                   protein describe in cmapfile. Used to calc    #
#                                                   ResType from ResNum                           #
#               $$cmap_hlist{"rlist"}       - Hash list of index to the shifts and                #
#                                             intensity and workbooks                             #
#               $cmap_row = $$cmap_hlist{"rlist"}{residuename}                                    #
#                                  residuename is the ResType + Index                             #
#                               $$cmap_row{"aa"}    - Residue Type of row (ie AA Type)            #
#                               $$cmap_row{"index"} - Absolute index starting from 1 of Current   #
#                                                     Residue                                     #
#                               $$cmap_row{"name"}  - ResType and Index concatenated              #
#                               $$cmap_row{"author_index"} - Sequence Number as per the Author /  #
#                                                            Creator of cmapfile                  #
#                               $$cmap_row{"author_name"}  - ResType and Sequence Number          #
#                                                           concatenated                          #
#                               $$cmap_row{"next"}    - Next residuename                          #
#                               $$cmap_row{"prev"}    - Previous residuename                      #
#                               $$cmap_row{"shifts"}  - Hash of shifts                            #
#                               $$cmap_row{"shifts"}{$shift_name}{"list"} - array of shift values #
#                                                                                                 #
# Purpose : To parse the input cmapfile and return the described data structure                   #
###################################################################################################

sub parseCMapfile
   { #===== begin parseCMapfile =====#
     
     # input cmap file
     my $cmap_infile = shift @_;
     
     my @input = ();
     my $tokens = ();   
     my $parsed_tokens = [];
     my $sections = [];
     my $sequence=();
     my $name_array=();
     my $cmap_hlist = {};

     # empty the comments element hash
     $$cmap_hlist{"comments"} = '';
     # empty the settings element hash
     $$cmap_hlist{"settings"} = [];

     # open the input cmap file or die
     local *IN;
     if ($cmap_infile =~ /^\-$/)
     {
	 *IN = *STDIN;
     }
     else
     {
	 open(IN, "<$cmap_infile") || die("Unable to open input file: " . $cmap_infile) ;
     }

     # While you still have lines.  Either push onto the input hash or the comment element of the hash
     while (my $line = <IN>)
     	{ 
	  if ($line =~ /^\#/)
	    {$$cmap_hlist{"comments"} .= $line;}
	  elsif ($line =~ /^Settings\:/i)
	    { push @{$$cmap_hlist{"settings"}}, $line; }
	  elsif ($line =~ /^\s*$/)
	    {}
	  else
	    {push (@input, $line);} 
	}
     # close the cmap file or die
     close(IN)  || die("Unable to close input file: " . $cmap_infile);
     
     # save the filename in the hash
     $$cmap_hlist{"filename"} = $cmap_infile;
     

     # WARNING CHEEZY CODE AHEAD ...
     # make a really big minimum so that it will be replaced.
     my $min_cmap_index = 100000; 

     my $not_first_time_through = -1;
     my $index = 0;
     my $aa = '';
     my $prev_residuename = '';
     my $residuename = '';
     my $residuenumber = '';
     my $name_conversion_hash = {};

     # while you have input lines in your input array do
     while (@input > 0)
       {
	 # get the next line
	 my $line = shift @input;
	 
	 # if the line starts with Assignment_Info, then the tokens will be on the next line
	 if ($line =~ /^Assignment_Info/)
	   {
	     $line = shift @input; 
	     @$tokens = split(/\s+/, $line);
	     shift @$tokens; # shift the AA token off

	     # add the new tokens to the parsed array
	     push @$parsed_tokens, @$tokens;
	     push @$sections, [ @$tokens ];

	     # reset the index
	     $index = 0;
	     $not_first_time_through++;
	   }
	 # otherwise if the line is not just spaces or empty then do ...
	 elsif ($line !~ /^\s*$/)
	   {
	     # add one to the index for the nonempty lines 
	     $index++;

	     # split the values by the spaces
	     my @chemical_shift_values = split(/\s+/, $line);


#	     for (my $count=0; $count < @chemical_shift_values; $count++)
#	       {
#		 print $chemical_shift_values[$count], "\n";
#	       }


	     # We assume that the first value is the residue
	     $prev_residuename = $residuename;
	     $residuenumber = shift @chemical_shift_values;

	     # If it is the first time through then set the minimum cmap index and spud the sequence.
	     $aa = substr($residuenumber,0,1);
	     $min_cmap_index =  ( substr($residuenumber,1) < $min_cmap_index ) ? substr($residuenumber,1) : $min_cmap_index;  
	     if ($not_first_time_through == 0)
	       {
	       push @$sequence, $aa;
	       }


	     
	     # Set the parenthesis initial position
	     my $paren_pos = 0;
	     
	     # for loop for the chemical shift values to look for parenthesis that got broken up
	     # this allows the token count to remain reasonable.
	     for (my $count=0; $count <= $#chemical_shift_values; $count++)
	       {
		 # if the value starts with a parenthesis then mark the start
		 if ($chemical_shift_values[$count] =~ /\(/)
		   {
		     $paren_pos = $count;
#		     print "repos\n";
		   }
		 # otherwise if the value contains the end parenthesis then reconcatenate the stuff
		 elsif ($chemical_shift_values[$count] =~ /\)/)
		   {
		     # reset the list
		     my $list = '';

		     # push from the beginning of the parenthesis to the current count onto a list
		     for (my $x = $paren_pos; $x <= $count; $x++)
		       {
			 $list .= $chemical_shift_values[$x] . " ";
		       }

		     # remove the ending spaces
		     $list =~ s/\s+$//;
#		     print ">>", $list, "\n";
 
		     # replace the position of the first parenthesis with the list of values
		     $chemical_shift_values[$paren_pos] = $list;

		     # take out the list of values from the chemical shift array
		     splice (@chemical_shift_values, ($paren_pos + 1), ($count - $paren_pos));

		     # set the counter to the next position after the splicings
		     $count = $paren_pos;
		   }
	       }

	     # save the aminoacid, index, BMRB name, author's numbering and naming (based on numbering)
	     if ($not_first_time_through == 0)
	       {
	       $residuename = $aa . $index;
	       push @$name_array, $residuename;
	       $$name_conversion_hash{$residuenumber} = $residuename;
	       }
	     else
	       { $residuename = $$name_conversion_hash{$residuenumber}; }

	     $$cmap_hlist{"rlist"}{$residuename}{"aa"} = $aa;
	     $$cmap_hlist{"rlist"}{$residuename}{"index"} = $index;
	     $$cmap_hlist{"rlist"}{$residuename}{"name"} = $aa . $index;
	     $$cmap_hlist{"rlist"}{$residuename}{"author_index"} = substr($residuenumber,1);
	     $$cmap_hlist{"rlist"}{$residuename}{"author_name"} = $residuenumber;

	     if ($prev_residuename ne '' && ($not_first_time_through == 0))
	     {
		 $$cmap_hlist{"rlist"}{$prev_residuename}{"next"} = $residuename;
		 $$cmap_hlist{"rlist"}{$residuename}{"prev"} = $prev_residuename;
	     }

	     # foreach chemical shift value do the following
	     for (my $count=0; $count <= $#chemical_shift_values; $count++)
	       {

		 # if it is an empty field then make an empty array ref
		 if ($chemical_shift_values[$count] =~ /^\-+$/)
		   {
		     $$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$$tokens[$count]}{"list"} = [];
		   }
		 
		 # otherwise if it has parenthesis (ie multiple values) split them and save as array
		 elsif ($chemical_shift_values[$count] =~ /^\(.+?\)$/)
		   {
#		       print "MULT>> $residue \t", $$tokens[$count], "\t\t",$chemical_shift_values[$count],"\n";
		     # remove leading and trailing spaces
		     $chemical_shift_values[$count] =~ s/^\(\s*(.*?)\s*\)$/$1/;
		     
		     # split the chemical shift value on the array
		     my @shifts_initial = split (/\s+/, $chemical_shift_values[$count]);
		    
		     # exclude empty elements from the final array
		     my @shifts = grep (!/^\-+$/, @shifts_initial);

		     # set the list of 
		     $$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$$tokens[$count]}{"list"} = [ @shifts ];
		   }

		 # otherwise if there is only one element, make an array of one element
		 else
		   {
		     $$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$$tokens[$count]}{"list"} = [ $chemical_shift_values[$count] ];
		   }
	       }
	   }
	 # Turn off first time through boolean as it is no longer true
       }

     # push the starting position and sequence and the tokens
     $$cmap_hlist{"sequence_startpos"} = $min_cmap_index;
     $$cmap_hlist{"sequence"} = $sequence;
     $$cmap_hlist{"name_array"} = $name_array;
     $$cmap_hlist{"parsedtokens"} = $parsed_tokens;
     $$cmap_hlist{"sections"} = $sections;
     
     # return the cmap list
     return $cmap_hlist;
   }	       


#################################### reMultiply ##################################
#                                                                                #
# Input   : cmap_hlist as described above                                        #
# Output  : corrected number of elements in the cmap_hlist                       #
# Purpose : to allow the output module to output correct number of blanks for aa #
##################################################################################

sub reMultiply
  {
      my $cmap_hlist = shift @_;

      my $prev_aa = '';
      foreach my $residuename ( @{$$cmap_hlist{"name_array"}} )
      {
	  my $current_aa = $$cmap_hlist{"rlist"}{$residuename}{"aa"};
	  
	  foreach my $token (@{$$cmap_hlist{"parsedtokens"}})
	  {

	      # GSS 06/17/2001   added if statement to avoid not ARRAY error in case no data 
	      #                  (bug in calling program)
	      if (!(exists $$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}))
	         {$$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}{"list"} = ["-"];}


	      my @temp_list;
	      if ($token =~ /^(.+?)(-1)$/)
	      {
		      if ($prev_aa ne '')
		      {
			  my $std_token = $1;

			  while (@{$$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}{"list"}} < $multiplicity{$prev_aa}{$std_token})
			  {
			      push @{$$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}{"list"}}, "";
			  }
		      }
		  }
		  else
		  {
		      if (exists $$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}{"list"})
		      {
			  while (@{$$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}{"list"}} < $multiplicity{$current_aa}{$token})
			  {
			      push @{$$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}{"list"}}, "";
			  }
		      }
		      else
		      {
			  @{$$cmap_hlist{"rlist"}{$residuename}{"shifts"}{$token}{"list"}} = ("-");
		      }
		  }
	      
	  }
	  $prev_aa = $current_aa; 
      }
  }

############################### sortResidueNumber ###############################
#                                                                               #
# Input   : The two AA# pairs                                                   #
# Output  : 1, 0, -1 depending on the comparison                                #
# Purpose : another small sorting subroutine for sorting by the residue number  #
#################################################################################

sub sortResidueNumber
  {
    my $c = $a;
    my $d = $b;

    # Get the residue number of A and B and compare them
    $c =~ s/.(.*?)/$1/;
    $d =~ s/.(.*?)/$1/;
    $c <=> $d 
  }

 
######################## outputCMapfile ################################
# Input   : cmap_hlist datatype described in parseCMapfile             # 
# Output  : a cmap file name determined by the cmap_hlist filename     #
# Purpose : To output an AutoAssign CMap Format Assignment file        #
########################################################################

sub outputCMapfile
  {
  # cmap data structure
  my $cmap_hlist = shift @_;
  my $filename = shift @_;
  my $write_options = {};
  if (@_)
    { $write_options = shift @_; }
  
  # readjust the number of dashes
  &reMultiply($cmap_hlist);
  
  # tokens to be used to output the files
  my $all_tokens = {};
  foreach my $token (@{$$cmap_hlist{"parsedtokens"}}) 
    { $$all_tokens{$token} = 1; }

  my $sections = [];
  if (exists $$cmap_hlist{"sections"} && (ref $$cmap_hlist{"sections"} eq "ARRAY"))
    { $sections = $$cmap_hlist{"sections"}; }
  
  # open the output file for writing and write the header.
  
  local *OUT;
  if ($filename =~ /^\-$/)
    { *OUT = *STDOUT; }
  else
    { open(OUT, ">". $filename) || die("Unable to open output file: " . $$cmap_hlist{"filename"}); }

  print OUT "\n";
  if (ref $$cmap_hlist{"settings"} eq "ARRAY" && (@{$$cmap_hlist{"settings"}} > 0))
    { 
    print OUT @{$$cmap_hlist{"settings"}}; 
    print OUT "\n";
    }
  
  my $collapse = 0;
  if ($$write_options{"collapse"} eq "all")
    { $collapse = 1; }
  
  my $used_tokens = {};
  foreach my $section_tokens (@$sections)
    { 
    &outputCMapSection($cmap_hlist, [ grep {! $$used_tokens{$_} && $$all_tokens{$_}; } (@$section_tokens) ], $collapse);
    if ($$write_options{"collapse"} ne "none")
      { $collapse = 1; }
    
    foreach my $token (@$section_tokens) 
      { $$used_tokens{$token} = 1; }
    } 
  
  if (! $$write_options{"sections_only"})
    { &outputCMapSection($cmap_hlist, [ grep {! $$used_tokens{$_} && $$all_tokens{$_}; } (keys %$all_tokens) ], $collapse); }
  
  
  # Close the file for output
  close(OUT) || die("Unable to close output file: " . $$cmap_hlist{"filename"});
  
  # return nothing, because the file was supposed to be output
  return;
  }			       


# outputCMapSection
#   Prints out a single CMap section
#
#   Parameters:
#	$cmap_hlist - ref to cmap hash structure
#	$tokens - ref to array of column tokens
#	$collapse - toggle for collapsing section (not printing residues with no values for given tokens).
#
sub outputCMapSection
  {
  my $cmap_hlist = shift @_;
  my $tokens = shift @_;
  my $collapse = shift @_;
  my $all_tokens = shift @_;


  # skip Section if it contains no tokens to print
  if (! @$tokens)
    { return; }

  print OUT "Assignment_Info\n";
  print OUT "AA\t", join("\t", @$tokens), "\n";
    
  # foreach element of the sorted residue number list do.  This makes sure that the cmap file is in order
  foreach my $resname (@{$$cmap_hlist{"name_array"}})
    {
    #Added for 0 and one previous from the first one.  Makes sure that is real, not just accessed once and thus existing.
    next if ($$cmap_hlist{"rlist"}{$resname}{"index"} eq '');
    
    my $line = "";
    my $value_found = 0;

    # add residue name
    if (defined $$cmap_hlist{"rlist"}{$resname}{"author_name"})
      { $line = $$cmap_hlist{"rlist"}{$resname}{"author_name"} . "\t"; }
    else
      { $line = $$cmap_hlist{"rlist"}{$resname}{"name"} . "\t"; }


    # forevery token, which is ordered as in the input datatype (arrays are ordered)
    foreach my $token (@$tokens)
      {
      # if the token type is an array then execute as follows (this is almost always followed)
      if ( exists $$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token} && $$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"} && ref($$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"}) eq "ARRAY")
	{
	my @temp_array = @{$$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"}};
	
	# If there are more than one element then put them in the parenthesis notation
	if (@temp_array > 1 )
	  {
	  # print the elements separated by a space.  Notice the last element's space is removed by the \b
	  # this is purely for aesthetic value.
	  
	  $line .= "(";
	  
	  # pull off the last element of the array
	  my $lastelem = pop @temp_array;
	  
	  # forevery element in the array, print it and a space.
	  foreach my $elem (@temp_array)
	    {
	    if ($elem !~ /^[-]*$/)
	      { 
	      $line .= $elem . " "; 
	      $value_found = 1;
	      }
	    else
	      { $line .= "- "; }
	    }
	  
	  # for the last element, print it out without the ending space
	  if ($lastelem !~ /^[-]*$/)
	    { 
	    $line .= $lastelem; 
	    $value_found = 1;
	    }
	  else
	    { $line .= "-"; }
	  
	  $line .= ")\t";
	  }
	# otherwise just print out the element or - if the element is null.
	else
	  {
	  if ($ {$$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"}}[0] !~ /^[-]*$/)
	    {
	    $line .= $ {$$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"}}[0] . "\t";
	    $value_found = 1;
	    }
	  else
	    { $line .= " - \t"; }
	  }
	}
      
      # otherwise not an array: THIS STATEMENT SHOULD NEVER BE EXECUTED in the current parsing format.
      # I have left it just as a remenant of the old way of doing things.
      else
	{
	if (exists $$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token} && $$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"} && $$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"} !~ /^[-]*$/)
	  {
	  $line .= $$cmap_hlist{"rlist"}{$resname}{"shifts"}{$token}{"list"} . "\t";
	  $value_found = 1;
	  }
	else
	  { $line .= " - \t"; }
	}
      }
    
    # print the line if collapse option is false or values where found.
    if (! $collapse || $value_found)
      { print OUT $line, "\n"; }

    }

  print OUT "\n\n";
  return;
  }



####################################### deMultiplicate ########################################
# Input   : a cmap_hlist type                                                                 # 
# Output  : a cmap_hlist type                                                                 #
# Purpose : To return only the proper multiplicity of the cmap_hlist                          #
###############################################################################################
sub deMultiplicate
   {
   my $cmap_hlist = shift @_;
   foreach my $residue_name (@{$$cmap_hlist{"name_array"}})
     {
     foreach my $shift_name (keys %{$$cmap_hlist{"rlist"}{$residue_name}{"shifts"}})
       {
       if ( exists $multiplicity{$$cmap_hlist{"rlist"}{$residue_name}{"aa"}}{$shift_name} )
	 { splice(@{$$cmap_hlist{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"list"}}, $multiplicity{$$cmap_hlist{"rlist"}{$residue_name}{"aa"}}{$shift_name}); }
       }
     }
   }


# 
# BMRB to CMap conversion routines
#

# readBMRBasCMAP
#   Reads a bmrb file in a CMap naming scheme and returns a BMRB/CMap hash of records. 
#
#   Parameters:
#       $input_bmrb_filename - bmrb filename to read
#       $entity_id - entity to convert; -1 to convert first one (optional).
#	$read_options - ref to read_options hash (optional). 
sub readBMRBasCMap
  {
  my $input_bmrb_filename = shift @_;
  my $entity_id = -1;
  if (@_)
    { $entity_id = shift @_; }
  my $read_options = {};
  if (@_)
    { $read_options = shift @_; }

  my $bmrb_hlist = &read_bmrb_file($input_bmrb_filename, $read_options);

  return &convertBMRBtoCMap($bmrb_hlist, $entity_id, $read_options);
  }


# convertBMRBtoCMap
#   Returns a CMap hash structure given a bmrb hash structure.
#
#
#
sub convertBMRBtoCMap
  {
  my $bmrb_hlist = shift @_;
  my $entity_id = -1;
  if (@_)
    { $entity_id = shift @_; }
  my $conversion_options = {};
  if (@_)
    { $conversion_options = shift @_; }

  if ($entity_id < 0)
    {
    if (! exists $$bmrb_hlist{"entity_list"} || ref($$bmrb_hlist{"entity_list"}) ne "HASH" || ! %{$$bmrb_hlist{"entity_list"}})
      {
      print STDERR "BMRB structure has no entity save_frames\n";
      return 0;
      }

    my @junk = sort { $a <=> $b; } (keys %{$$bmrb_hlist{"entity_list"}});
    $entity_id = $junk[0];
    }
  elsif (! exists $$bmrb_hlist{"entity_list"} || ref($$bmrb_hlist{"entity_list"}) ne "HASH" || ! exists $$bmrb_hlist{"entity_list"}{$entity_id})
    {
    print STDERR "BMRB structure does not have the \"$entity_id\" entity save_frame\n";
    return 0;
    }    

  my $entity = $$bmrb_hlist{"entity_list"}{$entity_id};

  # defining residue information
  my $cmap_hlist = {};
  @{$$cmap_hlist{"name_array"}} = @{$$entity{"name_array"}};
  $$cmap_hlist{"sequence"} = [ split(//,$$entity{"sequence"}) ];

  foreach my $residue_name (@{$$cmap_hlist{"name_array"}})
    {
    $$cmap_hlist{"rlist"}{$residue_name}{"name"}  = $$entity{"rlist"}{$residue_name}{"name"};
    $$cmap_hlist{"rlist"}{$residue_name}{"index"} = $$entity{"rlist"}{$residue_name}{"index"};
    if (exists $$conversion_options{"use_author_indexing"} && $$conversion_options{"use_author_indexing"})
      {
      $$cmap_hlist{"rlist"}{$residue_name}{"author_name"}  = $$entity{"rlist"}{$residue_name}{"author_name"};
      $$cmap_hlist{"rlist"}{$residue_name}{"author_index"} = $$entity{"rlist"}{$residue_name}{"author_index"};
      }
    else
      {
      $$cmap_hlist{"rlist"}{$residue_name}{"author_name"}  = $$entity{"rlist"}{$residue_name}{"name"};
      $$cmap_hlist{"rlist"}{$residue_name}{"author_index"} = $$entity{"rlist"}{$residue_name}{"index"};
      }
    $$cmap_hlist{"rlist"}{$residue_name}{"aa"}    = $$entity{"rlist"}{$residue_name}{"aa"};
    $$cmap_hlist{"rlist"}{$residue_name}{"prev"}  = $$entity{"rlist"}{$residue_name}{"prev"};
    $$cmap_hlist{"rlist"}{$residue_name}{"next"}  = $$entity{"rlist"}{$residue_name}{"next"};
    }
  
  my $peak_list_tokens = &convert_peak_lists($bmrb_hlist, $entity, $cmap_hlist);
  my $coupling_constants_tokens = &convert_coupling_constants($bmrb_hlist, $entity, $cmap_hlist);
  my $secondary_structure_tokens = &convert_secondary_structure($bmrb_hlist, $entity, $cmap_hlist);

  if (exists $$conversion_options{"interpret_atoms"} && ref($$conversion_options{"interpret_atoms"}) eq "ARRAY" && @{$$conversion_options{"interpret_atoms"}})
    { 
    &convert_assigned_chemical_shift_lists($entity,$cmap_hlist,$$conversion_options{"interpret_atoms"}, $$conversion_options{"interpret_non_HN"}); 
    push @$peak_list_tokens, @{$$conversion_options{"interpret_atoms"}};
    }
  
  my %seen;
  push @{$$cmap_hlist{"sections"}}, [grep {!$seen{$_}++} @$peak_list_tokens ];
  push @{$$cmap_hlist{"sections"}}, [grep {!$seen{$_}++} @$coupling_constants_tokens];
  push @{$$cmap_hlist{"sections"}}, [grep {!$seen{$_}++} @$secondary_structure_tokens];

  foreach my $cmap_hlist__sections (@{$$cmap_hlist{"sections"}})
    { push @{$$cmap_hlist{"parsedtokens"}}, @$cmap_hlist__sections; }

  return $cmap_hlist;
  }
  

#
# Internal routines
#

sub convert_peak_lists
  {
  my $bmrb_hlist = shift @_;
  my $entity = shift @_;
  my $cmap_hlist = shift @_;

  if (! exists $$bmrb_hlist{"peak_lists"} || ref($$bmrb_hlist{"peak_lists"}) ne "HASH" || ! %{$$bmrb_hlist{"peak_lists"}}) 
    { return []; }

  my $tokens     = [];
  my $seen       = {};
  foreach my $peak_list (values %{$$bmrb_hlist{"peak_lists"}})
    {
    foreach my $peak ( values %{$$peak_list{"plist"}} )
      {
      my $basisResidueIndex = -10000;
      my $basisResidueName;
      # count through the dimensions

      foreach my $dim (keys %{$$peak{"dimensions"}})
	{
	my $dimension = $$peak{"dimensions"}{$dim};
	if (exists $$dimension{"assignments"})
	  {
	  foreach my $assignment (@{$$dimension{"assignments"}})
	    {
	    if ($$assignment{"index"} > $basisResidueIndex)
	      { 
	      $basisResidueIndex = $$assignment{"index"}; 
	      $basisResidueName = $$assignment{"aa"} . $$assignment{"index"};
	      }
	    }
	  }
	}

      foreach my $dim (keys %{$$peak{"dimensions"}})
	{
	my $dimension = $$peak{"dimensions"}{$dim};
	if (exists $$dimension{"assignments"})
	  {
	  foreach my $assignment (@{$$dimension{"assignments"}})
	    {
	    my $residue_name   = $$assignment{"aa"} . $$assignment{"index"};
	    my $atom_name      = $$assignment{"atom_name"};
	    my $aa             = $$assignment{"aa"};
	    my $relative_index = $$assignment{"index"} - $basisResidueIndex;
	    my $suffix = "";

	    if ($relative_index < 0)
	      {$suffix = "-" . abs($relative_index);}
	    elsif ($relative_index > 0)
	      {$suffix = "+" . abs($relative_index);}
	    	    
	    if (! $$seen{$basisResidueName . "_" .  $atom_name . $suffix}++ )
	      {
	      my $cmap_atom_name = &convertAtomNameBMRBtoCMap($aa, $atom_name);
	      push @$tokens, "$cmap_atom_name$suffix";
	      if (exists $$entity{"rlist"}{$residue_name}{"shifts"}{$atom_name}{"idlist"})
		{
		my $list_count = 0;
		foreach my $shift_id (@{$$entity{"rlist"}{$residue_name}{"shifts"}{$atom_name}{"idlist"}})
		  {
		  if ($$assignment{"shift_id"} == $shift_id)
		    { push @ {$$cmap_hlist{"rlist"}{$basisResidueName}{"shifts"}{$cmap_atom_name . $suffix}{"list"}}, $$entity{"rlist"}{$residue_name}{"shifts"}{$atom_name}{"list"}[$list_count]; }

		  $list_count++;
		  if ($list_count >= @{$$entity{"rlist"}{$residue_name}{"shifts"}{$atom_name}{"list"}})
		    { $list_count = 0; }
		  }
		}
	      else
		{ print STDERR "Peak Assignment for $residue_name :: $atom_name inconsistent\n"; }
	      }
	    }
	  }
	}
      }
    }
  
  return $tokens;
  }


sub convert_assigned_chemical_shift_lists
  {
  my $entity = shift @_;
  my $cmap_hlist = shift @_;
  my $tokens     = shift @_;
  my $non_HN     = shift @_;

  foreach my $residue (values %{$$entity{"rlist"}})
    {
    next if (! $non_HN && ! (exists $$residue{"shifts"}{"H"} && @{$$residue{"shifts"}{"H"}{"list"}}) && ! (exists $$residue{"shifts"}{"N"} && @{$$residue{"shifts"}{"N"}{"list"}}));

    my $cmap_res_name = $$residue{"aa"} . $$residue{"index"};
    foreach my $atom_name (@$tokens)
      {
      next if (exists $$cmap_hlist{"rlist"}{$cmap_res_name} && exists $$cmap_hlist{"rlist"}{$cmap_res_name}{"shifts"}{$atom_name} 
	       && @{$$cmap_hlist{"rlist"}{$cmap_res_name}{"shifts"}{$atom_name}{"list"}});

      my $test_residue = $residue;
      my $base_atom_name = $atom_name;
      if ($atom_name =~ /\-1/)
	{
	next if (! exists $$residue{"prev"} || ($$residue{"prev"} eq ""));
	$test_residue = $$entity{"rlist"}{$$residue{"prev"}};
	$base_atom_name = substr($atom_name, 0,-2);
	}

      foreach my $shift_name (keys %{$$test_residue{"shifts"}})
	{
	next if (&convertAtomNameBMRBtoCMap($$test_residue{"aa"}, $shift_name) ne $base_atom_name);
	next if (! @{$$test_residue{"shifts"}{$shift_name}{"list"}});
	foreach my $value (@{$$test_residue{"shifts"}{$shift_name}{"list"}})
	  {
	  if (! (grep { $_ == $value; } (@{$$cmap_hlist{"rlist"}{$cmap_res_name}{"shifts"}{$atom_name}{"list"}})))
	    { push @{$$cmap_hlist{"rlist"}{$cmap_res_name}{"shifts"}{$atom_name}{"list"}}, $value; }
	  }
	}
      }
    }
  }

sub convert_secondary_structure
  {
  my $bmrb_hlist = shift @_;
  my $entity = shift @_;
  my $cmap_hlist = shift @_;
  
  if (! exists $$bmrb_hlist{"ss_lists"} || ref($$bmrb_hlist{"ss_lists"}) ne "HASH" || ! %{$$bmrb_hlist{"ss_lists"}}) 
    { return []; }
  
  my $CMAP_TOKEN = "S";
  foreach my $ss_list (values %{$$bmrb_hlist{"ss_lists"}})
    {
    foreach my $ss_elem (@{$$ss_list{"elements"}})
      {
      if ($$ss_elem{"entity_id"} == $$entity{"entity_id"})
	{
	for (my $count = $$ss_elem{"start_index"}; $count <= $$ss_elem{"end_index"}; $count++)
	  {
	  my $residue_name = &residueNameofIndex($$entity{"name_array"}, $count);
	  ${$$cmap_hlist{"rlist"}{$residue_name}{"shifts"}{$CMAP_TOKEN}{"list"}}[0] = &convertSSelemBMRBtoCMap($$ss_elem{"sseCode"});
	  } 
	}
      }
    }
  
  return [ $CMAP_TOKEN ];
  }


sub convert_coupling_constants
  {
  my $bmrb_hlist = shift @_;
  my $entity = shift @_;
  my $cmap_hlist = shift @_;

  if (! exists $$bmrb_hlist{"cc_lists"} || ref($$bmrb_hlist{"cc_lists"}) ne "HASH" || ! %{$$bmrb_hlist{"cc_lists"}}) 
    { return []; }

  my $tokens = [];
  foreach my $cc_list (values %{$$bmrb_hlist{"cc_lists"}})
    {
    foreach my $cc_elem (@{$$cc_list{"elements"}})
      {
      if ($$cc_elem{"Atom1entityID"} == $$entity{"entity_id"} && $$cc_elem{"Atom2entityID"} == $$entity{"entity_id"})
	{
	my $cmap_type = &convertJBMRBtoCMap($$cc_elem{"type"});
	if ($cmap_type ne "")
	  {
	  push @$tokens, $cmap_type;
	  my $residue_name = $$cc_elem{"Atom1residueName"} . $$cc_elem{"Atom1residueNumber"};
	  
	  if (($$cc_elem{"Atom2residueName"} eq "G")  && ($$cc_elem{"Atom2ambiguity"} ne "2"))
	    { $$cmap_hlist{"rlist"}{$residue_name}{"shifts"}{$cmap_type}{"list"}[0] = "i";}
	  else
	    { $$cmap_hlist{"rlist"}{$residue_name}{"shifts"}{$cmap_type}{"list"}[0] = &convert_jval_to_CMap($$cc_elem{"value"}, $$cc_elem{"error"}); }
	  }
	}
      }
    }

  return $tokens;
  }

# J VALUE STUFF
my $alpha = "6";
my $beta = "8";
my $granularIndefine = 1;

sub convert_jval_to_CMap
  {
  my $jval = shift @_;
  my $jerr = shift @_;
  
  if (($jval - $jerr <= $alpha) && ($jval + $jerr >= $beta))
    { return          "i"; }
  elsif (($jval-$jerr <= $alpha) && ($jval+$jerr <= $alpha))
    { return          "h"; }
  elsif (($jval-$jerr >= $beta) && ($jval+$jerr >= $beta))
    { return          "s"; }
  elsif (($jval-$jerr <= $alpha) && ($jval+$jerr < $beta) && $granularIndefine)
    { return          "hi"; }
  elsif (($jval-$jerr > $alpha) && ($jval+$jerr >= $beta) && $granularIndefine)
    { return          "si"; }
  else
    { return          "i"; }
  }


my %aa_name_conversion = ( "ala" => "A", "arg" => "R", "asn" => "N", "asp" => "D",
			   "cys" => "C", "gln" => "Q", "glu" => "E", "gly" => "G",
			   "his" => "H", "ile" => "I", "leu" => "L", "lys" => "K",
			   "met" => "M", "phe" => "F", "pro" => "P", "ser" => "S",
			   "thr" => "T", "trp" => "W", "tyr" => "Y", "val" => "V",
			   "a"   => "A", "r"   => "R", "n"   => "N", "d"   => "D", 
			   "c"   => "C", "q"   => "Q", "e"   => "E", "g"   => "G",
			   "h"   => "H", "i"   => "I", "l"   => "L", "k"   => "K",
			   "m"   => "M", "f"   => "F", "p"   => "P", "s"   => "S",
			   "t"   => "T", "w"   => "W", "y"   => "Y", "v"   => "V"); 

my %cmap_shift_conversion = 
  ( "A" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "C" => "CO", "CA" => "CA", "CB" => "CB", "N" => "N15", 
	       "HN" => "HN", "CO" => "CO", "N15" => "N15" },
    "R" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "HG2" => "HG",
	       "HG3" => "HG", "HD" => "HD", "HD2" => "HD", "HD3" => "HD", "HE" => "HE", "HH" => "HH", "HH1" => "HH", 
	       "HH11" => "HH", "HH2" => "HH", "HH12" => "HH", "HH21" => "HH", "HH22" => "HH", "C" => "CO", 
	       "CA" => "CA", "CB" => "CB", "CG" => "CG", "CD" => "CD", "CZ" => "CZ", "N" => "N15", "NE" => "NE", 
	       "NH" => "NH", "NH1" => "NH", "NH2" => "NH", "HN" => "HN", "CO" => "CO", "N15" => "N15" },
    "N" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HD2" => "HD", "HD21" => "HD",
	       "HD22" => "HD", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG" => "CG", "N" => "N15", "ND2" => "ND", 
	       "HN" => "HN", "CO" => "CO", "N15" => "N15", "ND" => "ND", "HD" => "HD" },
    "D" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "C" => "CO", "CA" => "CA",
	       "CB" => "CB", "CG" => "CG", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15", "OD1" => "OD",
	        "OD2" => "OD", "HD2" => "HD", "OD" => "OD", "HD" => "HD"},
    "C" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "C" => "CO",
	       "CA" => "CA", "CB" => "CB", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15", "SG" => "SG" },
    "B" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "C" => "CO",
	       "CA" => "CA", "CB" => "CB", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15", "SG" => "SG" },
    "Q" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "HG2" => "HG",
	       "HG3" => "HG", "HE2" => "HE", "HE21" => "HE", "HE22" => "HE", "C" => "CO", "CA" => "CA", "CB" => "CB",
	       "CG" => "CG", "CD" => "CD", "N" => "N15", "NE2" => "NE", "HN" => "HN", "CO" => "CO", "N15" => "N15",
	       "NE" => "NE", "HE" => "HE", "OE1" => "OE", "OE" => "OE"},
    "E" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "HG2" => "HG",
	       "HG3" => "HG", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG" => "CG", "CD" => "CD", "N" => "N15", 
	       "HN" => "HN", "CO" => "CO", "N15" => "N15", "OE1" => "OE", "OE2" => "OE", "OE" => "OE", "HE2" => "HE",
	       "HE" => "HE"},
    "G" => { "H" => "HN", "HA" => "HA", "HA2" => "HA", "HA3" => "HA", "C" => "CO", "CA" => "CA", "N" => "N15", 
	       "HN" => "HN", "CO" => "CO", "N15" => "N15" },
    "H" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HD1" => "HD", "HD2" => "HD",
	       "HE1" => "HE", "HE2" => "HE", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG" => "CG", "CD2" => "CD",
	       "CE1" => "CE", "N" => "N15", "ND1" => "ND", "NE2" => "NE", "ND" => "ND", "NE" => "NE", "CD" => "CD", 
	       "HN" => "HN", "CO" => "CO", "N15" => "N15", "CE" => "CE", "HD" => "HD", "HE" => "HE" },
    "I" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HG1" => "HG", "HG12" => "HG", "HG13" => "HG", 
	       "HG2" => "HG", "HD1" => "HD", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG1" => "CG", "CG2" => "CG",
	       "CD1" => "CD", "N" => "N15", "HD" => "HD", "HN" => "HN", "CO" => "CO", "N15" => "N15", "HG" => "HG",
	       "CD" => "CD", "CG" => "CG" },
    "L" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "HD" => "HD",
	       "HD1" => "HD", "HD2" => "HD", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG" => "CG", "CD" => "CD",
	       "CD1" => "CD", "CD2" => "CD", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15"},
    "K" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "HG2" => "HG",
	       "HG3" => "HG", "HD" => "HD", "HD2" => "HD", "HD3" => "HD", "HE" => "HE", "HE2" => "HE", "HE3" => "HE",
	       "HZ" => "HZ", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG" => "CG", "CD" => "CD", "CE" => "CE",
	       "N" => "N15", "NZ" => "NZ", "HN" => "HN", "CO" => "CO", "N15" => "N15" },
    "M" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "HG2" => "HG",
	       "HG3" => "HG", "HE" => "HE", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG" => "CG", "CE" => "CE",
	       "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15" , "SD" => "SD"},
    "F" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HD" => "HD", "HD1" => "HD",
	       "HD2" => "HD", "HE" => "HE", "HE1" => "HE", "HE2" => "HE", "HZ" => "HZ", "C" => "CO", "CA" => "CA", "CB" => "CB",
	       "CG" => "CG", "CD" => "CD", "CD1" => "CD", "CD2" => "CD", "CE1" => "CE", "CE2" => "CE", "CZ" => "CZ",
	       "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15" },
    "P" => { "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "HG2" => "HG", "HG3" => "HG",
	       "HD" => "HD", "HD2" => "HD", "HD3" => "HD", "C" => "CO", "CA" => "CA", "CB" => "CB", "CG" => "CG",
	       "CD" => "CD", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15" },
    "S" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HG" => "HG", "C" => "CO",
	       "CA" => "CA", "CB" => "CB", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15", "OG" => "OG" },
    "T" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HG1" => "HG", "HG2" => "HG", "C" => "CO", "CA" => "CA",
	       "CB" => "CB", "CG2" => "CG", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15", "CG" => "CG",
	       "HG" => "HG" },
    "W" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HD1" => "HD", "HE1" => "HE",
	       "HE3" => "HE", "HZ2" => "HZ", "HZ3" => "HZ", "HH2" => "HH", "C" => "CO", "CA" => "CA", "CB" => "CB",
	       "CG" => "CG", "CD1" => "CD", "CD2" => "CD", "CE2" => "CE", "CE3" => "CE", "CZ2" => "CZ",
	       "CZ3" => "CZ", "CH2" => "CH", "N" => "N15", "NE1" => "NE", "CZ" => "CZ", "HZ" => "HZ", "CD" => "CD",
	       "HN" => "HN", "CO" => "CO", "N15" => "N15", "HE" => "HE", "HH" => "HH", "NE" => "NE", "CE" => "CE",
	       "HD" => "HD", "CH" => "CH" },
    "Y" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HB2" => "HB", "HB3" => "HB", "HD" => "HD", "HD1" => "HD",
	       "HD2" => "HD", "HE" => "HE", "HE1" => "HE", "HE2" => "HE", "HH" => "HH", "C" => "CO", "CA" => "CA",
	       "CB" => "CB", "CG" => "CG", "CD" => "CD", "CD1" => "CD", "CD2" => "CD", "CE" => "CE", "CE1" => "CE",
	       "CE2" => "CE", "CZ" => "CZ", "N" => "N15", "HN" => "HN", "CO" => "CO", "N15" => "N15", "OH" => "OH" },
    "V" => { "H" => "HN", "HA" => "HA", "HB" => "HB", "HG" => "HG", "HG1" => "HG", "HG2" => "HG", "C" => "CO",
	       "CA" => "CA", "CB" => "CB", "CG" => "CG", "CG1" => "CG", "CG2" => "CG", "N" => "N15", 
	       "HN" => "HN", "CO" => "CO", "N15" => "N15" } );

################################# convertAtomNameBMRBtoCMap ###################################
# Input   : an amino acid and an BMRB atom name                                               # 
# Output  : a CMap atom name                                                                  #
# Purpose : To convert a BMRB atom name to its equivalent CMap atom name                      #
###############################################################################################
sub convertAtomNameBMRBtoCMap
  {
  my $aa        = shift @_;
  my $atom_name = shift @_;
  
  return $cmap_shift_conversion{$aa_name_conversion{lc($aa)}}{$atom_name};
  }

my %cmap_sse_conversion = 
    (
     "loop"        => "-",
     "alpha-helix" => "h",
     "turn"        => "t",
     "beta-sheet"  => "s"
     );

################################# convertSSelemBMRBtoCMap #####################################
# Input   : a BMRB secondary structure element code                                           # 
# Output  : a CMap secondary structure element code                                           #
# Purpose : To convert a BMRB SSE code to its equivalent CMap SSE code                        #
###############################################################################################
sub convertSSelemBMRBtoCMap
  {
  my $bmrb_sse        = shift @_;
  
  return $cmap_sse_conversion{lc($bmrb_sse)};
  }

my %cmap_J_conversion = 
    (
     "3JHNHA"        => "J3HNHA", 
     "J3HNHA" => "J3HNHA"
     );

#################################### convertJBMRBtoCMap #######################################
# Input   : a BMRB J coupling constant code                                                   # 
# Output  : a CMap J coupling constant code                                                   #
# Purpose : To convert a BMRB J code to its equivalent CMap J code                            #
###############################################################################################
sub convertJBMRBtoCMap
  {
  my $bmrb_J        = shift @_;
  
  return $cmap_J_conversion{uc($bmrb_J)};
  }

# module end
return 1;
