 package SparkyPeakParsing;

#
#   Author:     Gurmukh Sahota 06/15/2001
#   Copyright:  Gurmukh Sahota, 2001. All rights reserved.
#
#

# Returns a reference to a Array     of  a Hash      of an Array
#                          Residue  :::  AtomTypes  :::  possible chemical shifts

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw( parseSparkyPeakfile );
%EXPORT_TAGS = ( ALL => [@EXPORT_OK] );                                                   

# allows perl to fine local modules.
use FindBin;
use lib $FindBin::Bin;

use strict;


########################################## parsePeakfile #########################################################
# Input   : filename and tokens ie. dimensions used. Explained below                                             #
# Output  : A reference to a peakfile hash                                                                       #
#                $sparky_peakfile_hlist - reference to a hash                                                    #
#               $$sparky_peakfile_hlist{"filename"}    - peaklist's filename                                     #
#               $$sparky_peakfile_hlist{"dimlist"}      - the tokens of the peaklist or the dimension            #
#                                                 names (HN, N15, CX)                                            #
#               $$sparky_peakfile_hlist{"internal_comments"}    - the comments in the peaklist, lines            #
#                                                 starting with a #                                              #
#               $$sparky_peakfile_hlist{"end_comments"} - The filehistory created by some of the                 #
#                                                 AutoAssign Scripts                                             #
#               $$sparky_peakfile_hlist{"plist"}       - Hash list of index to the shifts and                    #
#                                                 intensity and workbooks                                        #
#               $peakfile_row = $$sparky_peakfile_hlist{"plist"}{indexnumber}                                    #
#                                                                                                                #
#                               $$peakfile_row{"shifts"}{token}               - shift value of the token         #
#                               $$peakfile_row{"intensity"}                   - intensity of the peak            #
#                               $$peakfile_row{"workbookname"}                - workbook name of the peak        #
#                               $$peakfile_row{"assign"}{"residue"}           - residue assignment of peak       #
#                               $$peakfile_row{"assign"}{"atomtype"}{"list"}  - atomType list for the assignment #
#                               $$peakfile_row{"PRL"}{"list"}                 - Possible Residue List(array)     #
#                               $$peakfile_row{"PAL"}{"list"}                 - Possible AtomType List(array)    #
#                                                                                                                #
# Purpose : To parse the input peakfile and return the described data structure                                  #
##################################################################################################################

sub parseSparkyPeakfile

  { #===== begin parsePeakfile =====#
    my $infile = shift @_;

    my @input = ();
    my $count = 0;
    my $line;

    my $sparky_peakfile_hlist = {};

    # open the file for input
     local *IN;
     if ($infile =~ /^\-$/)
     {
	 *IN = *STDIN;
     }
     else
     {
	 open(IN, "<$infile");
     }
    # save the input parameters in the hash for future reference
    $$sparky_peakfile_hlist{"filename"} = $infile;

    
    my $header_line = <IN>;
    $header_line =~ s/^\s+//;
    $header_line =~ s/\s+$//;

    my @tokens = split /\s+/, $header_line;

    shift @tokens; # Assignment
    pop @tokens;   # Height
    pop @tokens;   # Data

    $$sparky_peakfile_hlist{"dimlist"} = \@tokens;

    # while you do not hit the eof do the following
    while ($line = <IN>)
      { 
	# save the comments and push the other lines on to an input array.
	if ($line =~ /^\s*$/)   {next;}
	else                    {push (@input, $line);} 
      }
    
    # close the file
    close(IN);
    my $index=1;
    # while there are elements left in the input array
    while (@input > 0)
      {	# pull off the next line
	$line = shift @input;
	$line =~ s/^\s+//;
	$line =~ s/\s+$//;
	# as long as the line is not empty or just spaces do
	if ($line !~ /^\s*$/)
	  {
	    # split the line on the spaces
	    my @chemical_shifts = split(/\s+/, $line);

	    # we assume that the last column is the intensity
	    $$sparky_peakfile_hlist{"plist"}{$index}{"intensity"} = pop @chemical_shifts;

	    my $max_shifts = scalar(@chemical_shifts);
	    # the token columns are stored
	    for (my $count=0; $count < $max_shifts; $count++)
	      {
		last   if ($count > $#tokens);
		$$sparky_peakfile_hlist{"plist"}{$index}{"shifts"}{$tokens[$#tokens - $count]} = pop @chemical_shifts;
#		print $index, " :: ", $$sparky_peakfile_hlist{"plist"}{$index}{"shifts"}{$tokens[$#tokens - $count]}, "\n";
	      }
	    # we assume that the remaining columns will be the workbook name
	    $$sparky_peakfile_hlist{"plist"}{$index}{"workbookname"} = join("", @chemical_shifts);

	    $index++;
	}
    }
    
    # returned the datastructure
    return $sparky_peakfile_hlist;

  } #===== end   parsePeakfile =====#			       



