#!/usr/bin/perl
#!/usr/bin/perl
# apply_shifting.pl
# 	Adjust peak shifts using the given shifting.
# 
#	Copyright Hunter Moseley, 1999. All rights reserved.
#	Written by Hunter Moseley 4/4/1999
#
#   usage: apply_shifting.pl in_peak_file new_workbook_name out_peak_file shift1 shift2 ...
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 	workbook_name	- string to use as workbook name or "-" to use old name.
# 	shift1 shift2...- dimension shifts to apply.
#
use strict;

# Print usage info
if ((@ARGV < 4) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "apply_shifting.pl\n";
  print "	Adjust peak shifts using the given shifting.\n";
  print "\n";
  print "  usage: apply_shifting.pl in_peak_file new_workbook_name out_peak_file shift1 shift2 ...\n";
  print "\n";
  print "	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print "	workbook_name	- string to use as workbook name or \"-\" to use old name.\n";
  print "	shift1 shift2...- dimension shifts to apply.\n";
  exit(1);
  }

# Read in comand line parameters
my $in_peak_file = $ARGV[0];
my $workbook_name = $ARGV[1];
my $out_peak_file = $ARGV[2];

my $x;
my @reference_alist;
for($x = 3; $x < @ARGV; $x++)
  { push @reference_alist, $ARGV[$x]; }


# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);
my $num_cols = @{$$peak_alist[0]};

# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x < $num_cols-2; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print Rereferenced Peaks
my $peak;
my $x = 1;
my $y;
foreach $peak (@$peak_alist)
  { 
  for($y=0; $y < @reference_alist; $y++)
    { $$peak[$y+1] += $reference_alist[$y]; }

  $$peak[0] = $x++; 
  $$peak[$num_cols-2] = "\t" . $$peak[$num_cols-2];
  if ($workbook_name eq "-")
    {
    $$peak[$num_cols-1] = ("\t" x (length($$peak[$num_cols-2]) < 9)) . $$peak[$num_cols-1];
    }
  else
    {
    my ($notes) = ($$peak[$num_cols-1] =~ /[^.]*[.](.*)/);
    $$peak[$num_cols-1] = ("\t" x (length($$peak[$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes; 
    }
  print OUTPUT_FILE join("\t",@$peak), "\n";
  }

# Print Comments
print OUTPUT_FILE "*\n";

my $line;
my $history_found = 0;
foreach $line (@$end_comments_alist)
  {
  if (! $history_found && ($line =~ /^File History:/))
    {
    print OUTPUT_FILE "File History: $out_peak_file \n\n";
    print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
    $history_found = 1;
    next;
    }
  
  print OUTPUT_FILE $line;  
  }

if (! $history_found)
  {
  print OUTPUT_FILE "\nFile History: $out_peak_file \n\n";
  print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
  }

close OUTPUT_FILE;


#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }
