#!/usr/bin/perl
#!/usr/bin/perl
#  bmrb2AutoStructure.pl
#	Converts bmrb format to AutoStructure Resonance List format.
#
#  usage: bmrb2AutoStructure.pl input_bmrb_file autostructure_rl [options]
#
#		bmrb_file - name of bmrb file or "-" to read from standard input.
#		autostructure_rl - name of AutoStructure resonance list file or "-" to write to standard output.
#		options:
#			-restrict shift_list - restrict conversion to the given quoted list of shift names.
#			-entity idnum - entity to use.
#			-author - use author indexing instead.
#
#  Copyright Hunter Moseley, 2005. All rights reserved.
#  Written by Hunter Moseley 10/14/2005
#
use strict;

# allows perl to fine local modules.
use FindBin;
use lib $FindBin::Bin;

# contains subroutines for reading and writing bmrb files.
use BMRBParsing qw(:ALL);

# print help info
if ((@ARGV < 2) || ($ARGV[0] =~ /^\-h/))
  {
  print STDERR "Error: not enough parameters given.\n\n" if (@ARGV < 2);
  print STDERR "  bmrb2AutoStructure.pl\n";
  print STDERR "	Converts bmrb format to AutoStructure Resonance List format.\n";
  print STDERR "\n";
  print STDERR "  usage: bmrb2AutoStructure.pl input_bmrb_file autostructure_rl [options]\n";
  print STDERR "\n";
  print STDERR "		bmrb_file - name of bmrb file or \"-\" to read from standard input.\n";
  print STDERR "		autostructure_rl - name of AutoStructure resonance list file or \"-\" to write to standard output.\n";
  print STDERR "		options:\n";
  print STDERR "			-restrict shift_list - restrict conversion to the given quoted list of shift names.\n";
  print STDERR "			-entity idnum - entity to use.\n";
  print STDERR "			-author - use author indexing instead.\n";
  exit(1);
  }

# read command line options.
my $input_bmrb_filename = shift @ARGV;
my $output_autostructure_rl_filename = shift @ARGV;
my $entity_id = 0;
my $use_author_indexing = 0;
my @shift_name_list;

while (@ARGV && ($ARGV[0] =~ /^\-/))
  {
  my $switch = shift @ARGV;

  if ($switch =~ /^\-e/)
    { $entity_id = shift @ARGV; }
  elsif ($switch =~ /^\-r/)
    { @shift_name_list = split(",",shift @ARGV); }
  elsif ($switch =~ /^\-a/)
    { $use_author_indexing = 1; }
  else
    { 
    print STDERR "INVALID OPTION: $switch\n\nExiting Program ...\n\n";  
    exit(1);
    }
  }


# read the bmrb file
my $bmrb_hlist = &read_bmrb_file($input_bmrb_filename, {});

# open Sparky RL file
local *AUTOSTRUCTURE_FILE;
if ($output_autostructure_rl_filename eq "-")
  { *AUTOSTRUCTURE_FILE = *STDOUT; }
else
  { open(AUTOSTRUCTURE_FILE, ">$output_autostructure_rl_filename") || die ("open " . $output_autostructure_rl_filename . " failed."); }


# cycle through each residue
my $index=1;
foreach my $entity (values %{$$bmrb_hlist{"entity_list"}})
  {
  next if ($entity_id && $entity_id != $$entity{"entity_id"});

  foreach my $residue_name (@{$$entity{"name_array"}})
    {
    foreach my $shift_name (&sort_shift_names(keys %{$$entity{"rlist"}{$residue_name}{"shifts"}}))
      {
      if (!scalar(@shift_name_list) || scalar(grep { $_ eq $shift_name; } @shift_name_list))
	{
	foreach my $shift_value (@{$$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"list"}})
	  {
	  print AUTOSTRUCTURE_FILE $index,"\t";
	  if (! $use_author_indexing)
	    { print AUTOSTRUCTURE_FILE $$entity{"rlist"}{$residue_name}{"index"}; }
	  else
	    { print AUTOSTRUCTURE_FILE $$entity{"rlist"}{$residue_name}{"author_index"}; }

	  print AUTOSTRUCTURE_FILE " \t",&convert_aa1to3($$entity{"rlist"}{$residue_name}{"aa"})," \t",$shift_name," \t", $$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"atom_type"},"\t",$shift_value," \t. \t";
	  if (exists $$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"ambiguity_code"})
	    { print AUTOSTRUCTURE_FILE $$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"ambiguity_code"},"\n"; }
	  else
	    { print AUTOSTRUCTURE_FILE "1\n"; }

	  $index++;
	  }
	}
      }
    }
  }

close AUTOSTRUCTURE_FILE;
