#!/usr/bin/perl
#!/usr/bin/perl
#  bmrb2SparkyRL.pl
#	Converts bmrb format to Sparky Resonance List format.
#
#  usage: bmrb2SparkyRL.pl input_bmrb_file output_sparky_rl [options]
#
#		bmrb_file - name of bmrb file or "-" to read from standard input.
#		sparky_rl - name of sparky RL file or "-" to write to standard output.
#		options:
#			-restrict shift_list - restrict conversion to the given quoted list of shift names.
#			-entity idnum - entity to use.
#
#  Copyright Hunter Moseley, 2000. All rights reserved.
#  Written by Hunter Moseley 2/20/2001
#
use strict;

# allows perl to fine local modules.
use FindBin;
use lib $FindBin::Bin;

# contains subroutines for reading and writing bmrb files.
use BMRBParsing qw(:ALL);

# print help info
if ((@ARGV < 2) || ($ARGV[0] =~ /^\-h/))
  {
  print STDERR "Error: not enough parameters given.\n\n" if (@ARGV < 2);
  print STDERR "bmrb2SparkyRL.pl\n";
  print STDERR "	Converts bmrb format to Sparky Resonance List format.\n";
  print STDERR "\n";
  print STDERR "  usage: bmrb2SparkyRL.pl input_bmrb_file output_sparky_rl [options]\n";
  print STDERR "\n";
  print STDERR "		bmrb_file - name of bmrb file or \"-\" to read from standard input.\n";
  print STDERR "		sparky_rl - name of sparky RL file or \"-\" to write to standard output.\n";
  print STDERR "		options:\n";
  print STDERR "			-restrict shift_list - restrict conversion to the given quoted list of shift names.\n";
  print STDERR "			-entity idnum - entity to use.\n";
  exit(1);
  }

# read command line options.
my $input_bmrb_filename = shift @ARGV;
my $output_sparky_rl_filename = shift @ARGV;
my $entity_id = 0;
my @shift_name_list;

while (@ARGV && ($ARGV[0] =~ /^\-/))
  {
  my $switch = shift @ARGV;

  if ($switch =~ /^\-e/)
    { $entity_id = shift @ARGV; }
  elsif ($switch =~ /^\-r/)
    { @shift_name_list = split(",",shift @ARGV); }
  else
    { 
    print STDERR "INVALID OPTION: $switch\n\nExiting Program ...\n\n";  
    exit(1);
    }
  }


# read the bmrb file
my $bmrb_hlist = &read_bmrb_file($input_bmrb_filename, {});

# open Sparky RL file
local *SPARKY_FILE;
if ($output_sparky_rl_filename eq "-")
  { *SPARKY_FILE = *STDOUT; }
else
  { open(SPARKY_FILE, ">$output_sparky_rl_filename") || die ("open " . $output_sparky_rl_filename . " failed."); }


# cycle through each residue
foreach my $entity (values %{$$bmrb_hlist{"entity_list"}})
  {
  next if ($entity_id && $entity_id != $$entity{"entity_id"});

  foreach my $residue_name (@{$$entity{"name_array"}})
    {
    my $name = $residue_name;
    if (exists $$entity{"rlist"}{$residue_name}{"author_name"})
      { $name = $$entity{"rlist"}{$residue_name}{"author_name"}; }

    foreach my $shift_name (keys %{$$entity{"rlist"}{$residue_name}{"shifts"}})
      {
      if (!scalar(@shift_name_list) || scalar(grep { $_ eq $shift_name; } @shift_name_list))
	{
	foreach my $shift_value (@{$$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"list"}})
	  {
	  print SPARKY_FILE $name,"\t",$shift_name,"\t",$$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"atom_type"},"\t",$shift_value,"\n";
	  }
	}
      }
    }
  }

close SPARKY_FILE;
