#!/usr/bin/perl
#!/usr/bin/perl
#  bmrb2cmap.pl
#	Converts bmrb format to cmap format.
#
#  usage: bmrb2cmap.pl input_bmrb_file output_cmap_file [options]
#
#		input_bmrb_file - bmrb file to read ("-" to read from standard input).
#		output_cmap_file - output file to write to ("-" to write to standard output).
#		options:
#			-author - use author indexing.
#			-entity id_num - id of entity to use from bmrb file.
#			-interpret atom_list - add following atom_list using assigned chemical shifts if no assigned peak is available. 
#				atom_list - a quoted comma separated list of atoms in cmap format (i.e. CA, CA-1, HN, N15, CO, etc.) 
#			-noroot - interpret assigned chemical shifts that are not rooted by HN and N15 values.  
#			-restrict saveframe_list - restrict conversion to the given list of saveframes.
#				saveframe_list - comma separated list of save frame types (i.e. spectral_peak_list, coupling_constants, secondary_structure).
#
#  Copyright Gurmukh Sahota, 2002. All rights reserved.
#  Written by Gurmukh Sahota 05/19/2002
#
#  Modified by Hunter Moseley 9/24/2002
#
#
#  Updated to use new BMRBParsing/CMapParsing modules.
#  Copyright Hunter Moseley 2005. All rights reserved.
#
use FindBin;                    
use lib $FindBin::Bin;    

use CMapParsing qw(:ALL);
use strict;

# print help info
if ((@ARGV < 2) || ($ARGV[0] =~ /-h/))
  {
  print STDERR "Error: not enough parameters given.\n\n" if (@ARGV < 2);
  print STDERR "bmrb2cmap.pl\n";
  print STDERR "	Converts bmrb format to cmap format.\n";
  print STDERR "\n";
  print STDERR "  usage: bmrb2cmap.pl input_bmrb_file output_cmap_file [options]\n";
  print STDERR "\n";
  print STDERR "		input_bmrb_file - bmrb file to read (\"-\" to read from standard input).\n";
  print STDERR "		output_cmap_file - output file to write to (\"-\" to write to standard output).\n";
  print STDERR "		options:\n";
  print STDERR "			-author - use author indexing.\n";
  print STDERR "			-entity id_num - id of entity to use from bmrb file.\n";
  print STDERR "			-interpret atom_list - add following atom_list using assigned chemical shifts if no assigned peak is available. \n";
  print STDERR "				atom_list - a quoted comma separated list of atoms in cmap format (i.e. CA, CA-1, HN, N15, CO, etc.) \n";
  print STDERR "			-noroot - interpret assigned chemical shifts that are not rooted by HN and N15 values. \n";
  print STDERR "			-restrict saveframe_list - restrict conversion to the given list of saveframes.\n";
  print STDERR "				saveframe_list - comma separated list of save frame types (i.e. spectral_peak_list, coupling_constants, secondary_structure).\n";
  print STDERR "\n";
  exit(1);
  }

my $bmrb_infile = shift @ARGV;
my $cmap_outfile = shift @ARGV;
my $entity_id = -1;

my $read_options = {};

while (@ARGV)
  { 
  my $switch = shift @ARGV; 
  # GSS 04/14/2001 in order to allow more options in the future
  if ($switch =~ /^\-r/)
    { $$read_options{"limit_save_frames"} = [ split(",",shift @ARGV), qw(assigned_chemical_shifts entity assembly) ]; }
  elsif ($switch =~ /^\-i/)
    { $$read_options{"interpret_atoms"} = [ grep { $_ ne ""; } (split(", ",shift @ARGV)) ]; }
  elsif ($switch =~ /^\-n/)
    { $$read_options{"interpret_non_HN"} = 1; }
  elsif ($switch =~ /^\-e/)
    { $entity_id = shift @ARGV; }
  elsif ($switch =~ /^\-a/)
    { $$read_options{"use_author_indexing"} = 1; }
  else
    {
    print STDERR "ERROR: Unknown option \"$switch\" used.";
    exit(1);
    }
  }

my $cmap_hlist = &readBMRBasCMap($bmrb_infile, $entity_id, $read_options);
&outputCMapfile($cmap_hlist, $cmap_outfile);
