#!/usr/bin/perl
#!/usr/bin/perl
#  bmrb_apply_shifting.pl
#	Apply atom type shifting to the shifts of a bmrb file
#
#  usage: bmrb_apply_shifting.pl input_filename output_filename shift_pair shift_pair ...
#
#	filename - bmrb filename or "-" for standard input/output.
#	shift_pair - (atom_type shift_value).
#		atom_type - type of atom (ie. C, N, H).
#		shift_value - amount to shift by.
#
#  Copyright Hunter Moseley, 2000. All rights reserved.
#  Written by Hunter Moseley 7/31/2000
#
#
#  Updated to use new BMRBParsing/CMapParsing modules.
#  Copyright Hunter Moseley 2005. All rights reserved.
#
use strict;

# allows perl to fine local modules.
use FindBin;
use lib $FindBin::Bin;

# contains subroutines for reading and writing bmrb files.
use BMRBParsing qw(:ALL);


# print help info.
if ((@ARGV < 4) || (@ARGV % 2) || ($ARGV[0] =~ /^-h/))
  {
  print STDERR "Error: not enough parameters given.\n\n" if (@ARGV < 4);
  print STDERR "Error: incorrect number of parameters given.\n\n" if ((@ARGV % 2) && (@ARGV > 4) && ($ARGV[0] !~ /^-h/));
  print STDERR "  bmrb_apply_shifting.pl\n";
  print STDERR "	Apply atom type shifting to the shifts of a bmrb file\n";
  print STDERR "\n";
  print STDERR "  usage: bmrb_apply_shifting.pl input_filename output_filename shift_pair shift_pair ...\n";
  print STDERR "\n";
  print STDERR "	filename - bmrb filename or \"-\" for standard input/output.\n";
  print STDERR "	shift_pair - (atom_type shift_value).\n";
  print STDERR "		atom_type - type of atom (ie. C, N, H).\n";
  print STDERR "		shift_value - amount to shift by.\n";
  exit(1);
  }

# read command line arguments.
my $input_filename = shift @ARGV;
my $output_filename = shift @ARGV;

my %shifting;
while(@ARGV)
  {
  my $atom_name = shift @ARGV;
  $shifting{$atom_name} = shift @ARGV;
  }

# read the bmrb file
my $input_bmrb_hlist = &read_bmrb_file($input_filename, { } );

# cycle through each residue
foreach my $entity (values %{$$input_bmrb_hlist{"entity_list"}})
  {
  foreach my $residue_name (keys %{$$entity{"rlist"}})
    {
    # cycle through each shift
    foreach my $shift_name (keys %{$$entity{"rlist"}{$residue_name}{"shifts"}})
      {
      my $atom_type = $$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"atom_type"};
      # cycle through each shift value.
      foreach my $value (@{$$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}{"list"}})
	{ $value += $shifting{$atom_type}; } # update the shift
      }
    }
  }

# write updated bmrb file
&write_bmrb_file($input_bmrb_hlist, $output_filename);
