#!/usr/bin/perl
#!/usr/bin/perl
#  bmrb_extract_shifts.pl
#	Extract a list of shifts from a bmrb file.
#
#  usage: bmrb_extract_shifts.pl input_bmrb_file output_bmrb_file inclusion_toggle shift_name shift_name ...
#
#	bmrb_file - name of bmrb file or "-" for standard input/output.
#	inclusion_toggle:
#		inc(lusive) - include those shift names that match list.
#		exc(lusive) - exclude those shift names that match list.		 
#	shift_name - name of a shift in bmrb naming convention.
# 
#  Copyright Hunter Moseley, 2000. All rights reserved.
#  Written by Hunter Moseley 7/31/2000
#
#
#  Updated to use new BMRBParsing/CMapParsing modules.
#  Copyright Hunter Moseley 2005. All rights reserved.
#
use strict;

# allows perl to fine local modules.
use FindBin;
use lib $FindBin::Bin;

# contains subroutines for reading and writing bmrb files.
use BMRBParsing qw(:ALL);


# print help info.
if ((@ARGV < 4) || ($ARGV[0] =~ /-h/))
  {
  print STDERR "Error: not enough parameters given.\n\n" if (@ARGV < 4);
  print STDERR "bmrb_extract_shifts.pl\n";
  print STDERR "	Extract a list of shifts from a bmrb file.\n";
  print STDERR "\n";
  print STDERR "  usage: bmrb_extract_shifts.pl input_bmrb_file output_bmrb_file inclusion_toggle shift_name shift_name ...\n";
  print STDERR "\n";
  print STDERR "	bmrb_file - name of bmrb file or \"-\" for standard input/output.\n";
  print STDERR "	inclusion_toggle:\n";
  print STDERR "		inc(lusive) - include those shift names that match list.\n";
  print STDERR "		exc(lusive) - exclude those shift names that match list.\n";
  print STDERR "	shift_name - name of a shift in bmrb naming convention.\n";
  exit(1);
  }

# read command line arguments.
my $input_filename = shift @ARGV;
my $output_filename = shift @ARGV;
my $inclusion_toggle = shift @ARGV;
my @shift_name_list = @ARGV;


my $bmrb_hlist = &read_bmrb_file($input_filename, {});

# cycle through each residue
foreach my $entity (values %{$$bmrb_hlist{"entity_list"}})
  {
  foreach my $residue_name (keys %{$$entity{"rlist"}})
    {
    # split shifts of a residue.
    foreach my $shift_name (keys %{$$entity{"rlist"}{$residue_name}{"shifts"}})
      {
      my $test_match = scalar(grep { $_ eq $shift_name; } @shift_name_list);
      if ( ($test_match && ($inclusion_toggle =~ /^e/)) || (!$test_match && !($inclusion_toggle =~ /^e/)))
	{ delete $$entity{"rlist"}{$residue_name}{"shifts"}{$shift_name}; }
      }
    }
  }

# print the split bmrb files
&write_bmrb_file($bmrb_hlist, $output_filename);
