#!/usr/bin/perl
#!/usr/local/bin/perl
# change_intensity.pl
# 	Extract peaks with an abs(intensity) greater than a threshhold.
# 
#	Copyright Hunter Moseley, 2001. All rights reserved.
#	Written by Hunter Moseley 3/7/2001
#
#   usage: change_intensity.pl in_peak_file change_toggle amount new_workbook_name out_peak_file
# 
# 	peak_file 	 - name of peak file or "-" for standard input/output.
#	change_toggle:
#		change	 - change the intensity to the given amount.
#		add	 - add the given amount to the intensity.
#		sign_add - add/subtract the given amount to the intensity depending on the sign of the intensity.
# 	amount		 - amount to change intensity by.
# 	workbook_name	 - string to use as workbook name.
#

use strict;

#
# Print usage info
if ((@ARGV < 5) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "change_intensity.pl\n";
  print "	Extract peaks with an abs(intensity) greater than a threshhold.\n";
  print "\n";
  print "  usage: change_intensity.pl in_peak_file change_toggle amount new_workbook_name out_peak_file\n";
  print "\n";
  print "	peak_file 	 - name of peak file or \"-\" for standard input/output.\n";
  print "	change_toggle:\n";
  print "		change	 - change the intensity to the given amount.\n";
  print "		add	 - add the given amount to the intensity.\n";
  print "		sign_add - add/subtract the given amount to the intensity depending on the sign of the intensity.\n";
  print " 	amount		 - amount to change intensity by.\n";
  print "	workbook_name	 - string to use as workbook name.\n";
  exit(1);
  }

# Read command line arguements.
my $in_peak_file = $ARGV[0];
my $change_toggle = $ARGV[1];
my $change_amount = $ARGV[2];
my $workbook_name = $ARGV[3];
my $out_peak_file = $ARGV[4];

if ($change_toggle !~ /^[cas]/)
  {
  print "\nError: invalid change type given.\n";
  print "\n";
  exit(1);
  }

# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);
my $num_cols = @{$$peak_alist[0]};

# Sort peaks in descending order

# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x < $num_cols-2; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print High Intensity Peaks
my $peak;
my $x = 1;
foreach $peak (@$peak_alist)
  { 
  if ($change_toggle =~ /^c/)
    { $$peak[$num_cols-2] = $change_amount; }
  elsif($change_toggle =~ /^a/)
    { $$peak[$num_cols-2] += $change_amount; }
  elsif($$peak[$num_cols-2] > 0)
    { $$peak[$num_cols-2] += $change_amount; }
  elsif($$peak[$num_cols-2] < 0)
    { $$peak[$num_cols-2] -= $change_amount; }

  $$peak[$num_cols-2] = "\t" . $$peak[$num_cols-2];
  my ($notes) = ($$peak[$num_cols-1] =~ /[^.]*[.](.*)/);
  $$peak[$num_cols-1] = ("\t" x (length($$peak[$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes; 
  print OUTPUT_FILE join("\t",@$peak), "\n";
  }

# Print Comments
print OUTPUT_FILE "*\n";

my $line;
my $history_found = 0;
foreach $line (@$end_comments_alist)
  {
  if (! $history_found && ($line =~ /^File History:/))
    {
    print OUTPUT_FILE "File History: $out_peak_file \n\n";
    print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
    $history_found = 1;
    next;
    }
  
  print OUTPUT_FILE $line;  
  }

if (! $history_found)
  {
  print OUTPUT_FILE "\nFile History: $out_peak_file \n\n";
  print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
  }

close OUTPUT_FILE;

#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

