#!/usr/bin/perl
#!/usr/bin/perl
#!/usr/bin/perl
# concatenate.pl 
# 	Concatenates multiple peak files into new peak file.
# 
#	Copyright Hunter Moseley, 2000. All rights reserved.
#	Written by Hunter Moseley 9/20/2000
#
#       Modified Gurmukh Sahota 06/17/2001
#            Allow use of - in new_workbook_name to signify to keep the workbook name from
#            the original peaklist.
#
#
#   usage: concatenate.pl out_peak_file new_workbook_name [options] in_peak_file in_peak_file ...
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 	workbook_name	- string to use as workbook name or "-" to use old name.
#	options:
#		-old_index - use old peak indexes.
#
use strict;

# Print usage info
if ((@ARGV < 4) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "concatenate.pl\n";
  print " 	Concatenates multiple peak files into new peak file.\n";
  print " \n";
  print "   usage: concatenate.pl out_peak_file new_workbook_name [options] in_peak_file in_peak_file ...\n";
  print "\n";
  print " 	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print " 	workbook_name	- string to use as workbook name or \"-\" to use old name.\n";
  print "	options:\n";
  print "		-old_index - use old peak indexes.\n";
  exit(1);
  }

# Read in comand line parameters
my $out_peak_file = $ARGV[0];
my $workbook_name = $ARGV[1];
my $use_old_index = 0;
my $in_file_start = 2;
if ($ARGV[2] =~ /^-o/)
  {
  $use_old_index = 1;
  $in_file_start = 3;
  }

my @in_peak_files;
for(my $x= $in_file_start; $x < @ARGV; $x++)
  { push @in_peak_files, $ARGV[$x]; }

# Read Peak files
my @peak_lists;
foreach my $peak_file (@in_peak_files)
  {
  my ($peak_alist, $end_commands_alist) = &read_peak_file($peak_file);
  push @peak_lists, [ $peak_alist, $end_commands_alist, $peak_file ];
  }

my $num_cols = @{$peak_lists[0][0][0]};


# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x < $num_cols-2; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print each peak file
my $index = 1;
foreach my $peak_list (@peak_lists)
  {
  foreach my $peak (@{$$peak_list[0]})
    {
    if (! $use_old_index)
      { $$peak[0] = $index++; }
    $$peak[$num_cols-2] = "\t" . $$peak[$num_cols-2];

    # GSS 06/17/2001   added if statement to allow for the original workbook names to be kept.
    if ($workbook_name  eq "-")
	{
	  $$peak[$num_cols-1] = ("\t" x (length($$peak[$num_cols-2]) < 9)) . $$peak[$num_cols-1];
	}
    else
	{
           my ($notes) = ($$peak[$num_cols-1] =~ /[^.]*[.](.*)/);
           $$peak[$num_cols-1] = ("\t" x (length($$peak[$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes;
	} 
    print OUTPUT_FILE join("\t",@$peak), "\n";
    }
  }


# Print Comments
print OUTPUT_FILE "*\n";
print OUTPUT_FILE "File History: $out_peak_file \n\n";
print OUTPUT_FILE $0, " ", join(' ',@ARGV), "\n\n";

foreach my $peak_list (@peak_lists)
  {
  print OUTPUT_FILE "\n\nComments: $$peak_list[2]\n";
  print OUTPUT_FILE "-" x 80, "\n";
  print OUTPUT_FILE @{$$peak_list[1]};
  }

close OUTPUT_FILE;


#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

