#!/usr/bin/perl
#!/usr/bin/perl
# create_peak_list.pl 
# 	Create a peak list given the shift values and intensity in separate columns.  
#	The script adds an index column to the front and destroys note columns and 
#	appends them to the workbook_name.
# 
#	Copyright Hunter Moseley, 1999. All rights reserved.
#	Written by Hunter Moseley 4/24/1999
#
#   usage: create_peak_list.pl in_peak_file new_workbook_name out_peak_file note1 note2 ...
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 	workbook_name	- string to use as workbook name.
# 	note1 note2 ... - columns that should be appended to the workbook_name as notes
# 			  and then deleted; range [1..].
# 
use strict;

# Print usage info
if ((@ARGV < 3) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "create_peak_list.pl \n";
  print " 	Create a peak list given the shift values and intensity in separate columns.\n";
  print "	The script adds an index column to the front and destroys note columns and\n";
  print "	appends them to the workbook_name.\n";
  print "\n";
  print "  usage: create_peak_list.pl in_peak_file new_workbook_name out_peak_file note1 note2 ...\n";
  print "\n";
  print "	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print "	workbook_name	- string to use as workbook name.\n";
  print "	note1 note2 ... - columns that should be appended to the workbook_name as notes \n";
  print "			  and then deleted; range [1..].\n";
  exit(1);
  }

# Read in command line parameters
my $input_peak_file = $ARGV[0];
my $workbook_name = $ARGV[1];
my $out_peak_file = $ARGV[2];
my $x;
my @note_alist;
my %note_hlist;
my $note_count = 0;
for($x = 3; $x < @ARGV; $x++)
  { 
  $note_count++;
  push @note_alist, $ARGV[$x]; 
  $note_hlist{$ARGV[$x]} = 1;
  }

# Read Peak file
my $peak_alist = &read_peak_file($input_peak_file);

my $num_cols;
my $peak;
LOOP1: foreach $peak (@$peak_alist) # find num columns
  {
  my $x;
  next LOOP1 if (! @$peak);

  for($x = 0; $x < @$peak; $x++)
    { next LOOP1 if  (! $note_hlist{$x+1} && $$peak[$x] !~/\d/); }
  
  $num_cols = @$peak;
  last;
  }

# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x <= ($num_cols - @note_alist - 1); $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print peaks with extracted/reorganized dimensions.
my $x = 1;
my $peak;
my $y;
my $line_num = -1;
LOOP2: foreach $peak (@$peak_alist)
  { 
  $line_num += 1;
  next LOOP2 if  (! @$peak);

  for($y = 0; $y < @$peak; $y++) # check for valid peak
    { 
    if  (! $note_hlist{$y+1} && $$peak[$y] !~/\d/)
      {
      print STDERR "Skipping Line ",$line_num+1," \"",join(" ",@$peak),"\"\n";
      next LOOP2;
      }
    }

  if (@$peak < ($num_cols - $note_count))
    {
    print STDERR "Skipping Line ",$line_num+1," \"",join(" ",@$peak),"\"\n";
    next LOOP2;
    }

  my @new_peak = ();
  push @new_peak, $x++; # add index number

  my $col;
  for($y = 0; $y < @$peak; $y++) # add dimensions and intensity
    {
    next if ($note_hlist{$y+1});
    push @new_peak, $$peak[$y];
    }
  $new_peak[$#new_peak] = "\t" . $new_peak[$#new_peak];

  my $local_workbook_name = $workbook_name; # create workbook name and notes
  my $total_note;
  foreach my $note (@note_alist)
    { 
    if (($note <= @$peak) && ($$peak[$note-1] ne ""))
      { $total_note .= "." . $$peak[$note-1]; }
    }

  if ($total_note ne "")
    { $local_workbook_name .= $total_note; }

  push @new_peak, ("\t" x (length($new_peak[$#new_peak]) < 9)) . $local_workbook_name; 

  print OUTPUT_FILE join("\t",@new_peak), "\n";
  }

# Print Comments
print OUTPUT_FILE "*\n";
print OUTPUT_FILE "File History: $out_peak_file \n\n";
print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";

close OUTPUT_FILE;

#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

