#!/usr/bin/perl
#!/usr/bin/perl
# extract_by_dup_count.pl 
# 	Extract certain groups of peaks with matching chemical shifts and a certain peak count.
# 
#	Copyright Hunter Moseley, 2000. All rights reserved.
#	Written by Hunter Moseley 5/11/2000
#
#   usage: extract_by_dup_count.pl in_peak_file inclusion_toggle count_qualifier \
#			           new_workbook_name out_peak_file to11 tol2 ...
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 	inclusion_toggle:
# 		inc(lusive) - include only peaks with count_qualifier.
# 		exc(lusive) - exclude peaks with count_qualifier.
#	count_qualifier - (comparison_operator peak_count).
#		comparison_operator - "equal","less","greater","less_equal","greater_equal".
#		peak_count          - number of peaks in the glob.
# 	workbook_name	- string to use as workbook name or "-" to use old name.
# 	tol1 tol2 ... 	- dimension tolerances for comparison or "0" not to use specific 
# 			  dimension.
#
use strict;

# Print usage info
if ((@ARGV < 6) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "Extract certain groups of peaks with matching chemical shifts and a certain peak count.\n";
  print "\n";
  print "  usage: extract_by_dup_count.pl in_peak_file count_qualifier new_workbook_name \\ \n";
  print "                                 out_peak_file to11 tol2 ...\n";
  print "\n";
  print "	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print "	count_qualifier - (comparison_operator peak_count).\n";
  print "		comparison_operator - \"equal\",\"less\",\"greater\",\"less_equal\",\"greater_equal\".\n";
  print "		peak_count          - number of peaks in the glob.\n";
  print "	workbook_name	- string to use as workbook name or \"-\" to use old name.\n";
  print "	tol1 tol2 ... 	- dimension tolerances for comparison or \"0\" not to use specific \n";
  print "			  dimension.\n";
  exit(1);
  }

# Read in comand line parameters
my $in_peak_file = $ARGV[0];
my $comparison_operator = $ARGV[1];
my $comparison_value = $ARGV[2];
my $workbook_name = $ARGV[3];
my $out_peak_file = $ARGV[4];

my $x;
my @tolerance_alist;
for($x = 5; $x < @ARGV; $x++)
  { push @tolerance_alist, $ARGV[$x]; }

# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);
my $num_cols = @{$$peak_alist[0]};

# Create Equivalence array
my @equiv_alist;
my $pindex1;
my $duplicate_alist;
my @peak_count;
for($pindex1 = 0; $pindex1 < @$peak_alist; $pindex1++)
  {
  if (! $equiv_alist[$pindex1])
    { 
    $equiv_alist[$pindex1] = $pindex1+1; 
    $peak_count[$pindex1] = 1;
    }
    
  my $duplicate_alist = &find_duplicate_index_list($pindex1, $peak_alist, \@tolerance_alist);
  if (@$duplicate_alist)
    {
    my $dup_index;
    foreach $dup_index (@$duplicate_alist)
      {
      if (! $equiv_alist[$dup_index])
	{ 
	$equiv_alist[$dup_index] = $equiv_alist[$pindex1]; 
	$peak_count[$equiv_alist[$pindex1]-1]++;
	}
      }
    }
  }


# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x < $num_cols-2; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print duplicate peak list.
my $pindex;
my $x = 1;
for($pindex = 0; $pindex < @$peak_alist; $pindex++)
  {
  if (&peak_count_test($peak_count[$pindex]) && $equiv_alist[$pindex]) # is the peak to be printed?
    {
    my $equiv_value = $equiv_alist[$pindex];
    my $pindex2;
    for($pindex2 = $pindex; $pindex2 < @$peak_alist; $pindex2++)
      {
      if ($equiv_alist[$pindex2] == $equiv_value)
	{
	$equiv_alist[$pindex2] = 0;
	$$peak_alist[$pindex2][0] = $x++; 
	$$peak_alist[$pindex2][$num_cols-2] = "\t" . $$peak_alist[$pindex2][$num_cols-2];
	if ($workbook_name eq "-")
	  {
	  $$peak_alist[$pindex2][$num_cols-1] = ("\t" x (length($$peak_alist[$pindex2][$num_cols-2]) < 9)) . $$peak_alist[$pindex2][$num_cols-1];
	  }
	else 
	  {
	  my ($notes) = ($$peak_alist[$pindex2][$num_cols-1] =~ /[^.]*[.](.*)/);
	  $$peak_alist[$pindex2][$num_cols-1] = ("\t" x (length($$peak_alist[$pindex2][$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes; 
	  }
	print OUTPUT_FILE join("\t",@{$$peak_alist[$pindex2]}), "\n";
	}
      }
      print OUTPUT_FILE "#\n"; 
    }
  elsif($equiv_alist[$pindex])
    {
    my $equiv_value = $equiv_alist[$pindex];
    my $pindex2;
    for($pindex2 = $pindex; $pindex2 < @$peak_alist; $pindex2++)
      {
      if ($equiv_alist[$pindex2] == $equiv_value)
	{
	$equiv_alist[$pindex2] = 0;
	}
      }
    }
  }

# Print Comments
print OUTPUT_FILE "*\n";

my $line;
my $history_found = 0;
foreach $line (@$end_comments_alist)
  {
  if (! $history_found && ($line =~ /^File History:/))
    {
    print OUTPUT_FILE "File History: $out_peak_file \n\n";
    print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
    $history_found = 1;
    next;
    }
  
  print OUTPUT_FILE $line;  
  }

if (! $history_found)
  {
  print OUTPUT_FILE "\nFile History: $out_peak_file \n\n";
  print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
  }

close OUTPUT_FILE;


#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

sub find_duplicate_index_list
  {
  my $pindex = shift @_;
  my $peak_alist = shift @_;
  my $tolerance_alist = shift @_;

  my $duplicate_alist = [];
  my $x;
  my $y;
  for($y = $pindex+1; $y < @$peak_alist; $y++)
    {
    for($x=0; $x < @$tolerance_alist; $x++)
      {
      if (($$tolerance_alist[$x] != 0.0) && (abs($$peak_alist[$pindex][$x+1] - $$peak_alist[$y][$x+1]) >= $$tolerance_alist[$x]))
	{ last; }
      }

    if ($x == @$tolerance_alist)
      { push @$duplicate_alist, $y; }
    }

  return $duplicate_alist;
  }

sub peak_count_test
  {
  my $peak_count = shift @_;
  
  if ( (($comparison_operator =~ /^eq/) && ($peak_count != $comparison_value)) ||	# equal
       (($comparison_operator =~ /^l.*\_e/) && ($peak_count > $comparison_value)) || # less_equal
       (($comparison_operator =~ /^g.*\_e/) && ($peak_count < $comparison_value)) ||	# greater_equal
       (($comparison_operator =~ /^l(?!.*\_e)/) && ($peak_count >= $comparison_value)) ||	# less
       (($comparison_operator =~ /^g(?!.*\_e)/) && ($peak_count <= $comparison_value)) )	# greater
    { return 0; }

  return 1;  
  }
