#!/usr/bin/perl
#!/usr/bin/perl
# extract_by_shift_range.pl
# 	Extract peaks with chemical shifts within a range of values.
# 
#	Copyright Hunter Moseley, 2002. All rights reserved.
#	Written by Hunter Moseley 9/5/2002
#
#   usage: extract_by_shift_range.pl in_peak_file new_workbook_name out_peak_file shift_range1 shift_range2 ... 
# 
# 	peak_file 	 - name of peak file or "-" for standard input/output.
# 	workbook_name	 - string to use as workbook name.
#	shift_range      - (lower_bound upper_bound) or (N N) for none.
#

use strict;

#
# Print usage info
#
if ((@ARGV < 5) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "extract_by_shift_range.pl\n";
  print "	Extract peaks with chemical shifts within a range of values.\n";
  print "\n";
  print "  usage: extract_by_shift_range.pl in_peak_file new_workbook_name out_peak_file shift_range1 shift_range2 ... \n";
  print "\n";
  print "	peak_file 	 - name of peak file or \"-\" for standard input/output.\n";
  print "	workbook_name	 - string to use as workbook name.\n";
  print "	shift_range      - (lower_bound upper_bound) or (N N) for none.\n";
  exit(1);
  }

#
# Read command line arguements.
#
my $in_peak_file = $ARGV[0];
my $workbook_name = $ARGV[1];
my $out_peak_file = $ARGV[2];
my $x = 3;
my @shift_ranges;
while ($x < @ARGV)
  {
  push @shift_ranges, [ $ARGV[$x], $ARGV[$x+1] ];
  $x += 2;
  }

#
# Read Peak file
#
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);
my $num_cols = @{$$peak_alist[0]};


#
# Open Output file
#
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

#
# Print Header
#
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x < $num_cols-2; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

#
# Print Peaks that pass the shift_test
#
my $peak;
my $x = 1;
foreach $peak (@$peak_alist)
  {
  if (&shift_test($peak))
    {
    $$peak[0] = $x++; 
    $$peak[$num_cols-2] = "\t" . $$peak[$num_cols-2];
    my ($notes) = ($$peak[$num_cols-1] =~ /[^.]*[.](.*)/);
    $$peak[$num_cols-1] = ("\t" x (length($$peak[$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes; 
    print OUTPUT_FILE join("\t",@$peak), "\n";
    }
  }

# Print Comments
print OUTPUT_FILE "*\n";

my $line;
my $history_found = 0;
foreach $line (@$end_comments_alist)
  {
  if (! $history_found && ($line =~ /^File History:/))
    {
    print OUTPUT_FILE "File History: $out_peak_file \n\n";
    print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
    $history_found = 1;
    next;
    }
  
  print OUTPUT_FILE $line;  
  }

if (! $history_found)
  {
  print OUTPUT_FILE "\nFile History: $out_peak_file \n\n";
  print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
  }

close OUTPUT_FILE;

#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }


sub shift_test
  {
  my $peak = shift @_;
  
  my $x;
  for($x=0; $x < @shift_ranges; $x++)
    {
    if (($shift_ranges[$x][0] !~ /^[Nn]/) && 
	(($$peak[$x+1] < $shift_ranges[$x][0]) || ($$peak[$x+1] > $shift_ranges[$x][1])))
      { return 0; }
    }

  return 1;  
  }
