#!/usr/bin/perl
#!/usr/bin/perl
# extract_dims.pl 
# 	Generate new peak file by extracting/reorganizing specific dimensions from 
# 	another peak file.
# 
#	Copyright Hunter Moseley, 1999. All rights reserved.
#	Written by Hunter Moseley 4/3/1999
#
#   usage: extract_dims.pl in_peak_file new_workbook_name out_peak_file dim1 dim2 ...
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 	workbook_name	- string to use as workbook name.
# 	dim1 dim2 ... 	- column numbers for dimensions to extract in the order they 
# 			  should appear; range [1..].
# 
use strict;

# Print usage info
if ((@ARGV < 4) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "extract_dims.pl \n";
  print "	Generate new peak file by extracting/reorganizing specific dimensions from \n";
  print "	another peak file.\n";
  print "\n";
  print "  usage: extract_dims.pl in_peak_file new_workbook_name out_peak_file dim1 dim2 ...\n";
  print "\n";
  print "	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print "	workbook_name	- string to use as workbook name.\n";
  print "	dim1 dim2 ... 	- column numbers for dimensions to extract in the order they \n";
  print "			  should appear; range [1..].\n";
  exit(1);
  }

# Read in command line parameters
my $input_peak_file = $ARGV[0];
my $workbook_name = $ARGV[1];
my $out_peak_file = $ARGV[2];
my $x;
my @dim_alist;
for($x = 3; $x < @ARGV; $x++)
  { push @dim_alist, $ARGV[$x]; }


# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($input_peak_file);
my $num_cols = @{$$peak_alist[0]};

# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x <= @dim_alist; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print peaks with extracted/reorganized dimensions.
my $x = 1;
my $peak;
foreach $peak (@$peak_alist)
  { 
  my @new_peak = ();
  push @new_peak, $x++;
  my $dim;
  foreach $dim (@dim_alist)
    { push @new_peak, $$peak[$dim]; }
  push @new_peak, "\t" . $$peak[$num_cols-2];
  my ($notes) = ($$peak[$num_cols-1] =~ /[^.]*[.](.*)/);
  push @new_peak, ("\t" x (length($$peak[$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes; 
  print OUTPUT_FILE join("\t",@new_peak), "\n";
  }

# Print Comments
print OUTPUT_FILE "*\n";

my $line;
my $history_found = 0;
foreach $line (@$end_comments_alist)
  {
  if (! $history_found && ($line =~ /^File History:/))
    {
    print OUTPUT_FILE "File History: $out_peak_file \n\n";
    print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
    $history_found = 1;
    next;
    }
  
  print OUTPUT_FILE $line;  
  }

if (! $history_found)
  {
  print OUTPUT_FILE "\nFile History: $out_peak_file \n\n";
  print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
  }

close OUTPUT_FILE;

#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

