#!/usr/bin/perl
#!/usr/bin/perl
#!/usr/local/bin/perl
# extract_duplicates.pl 
# 	Extract a group of peaks that have duplicate chemical shifts.
# 
#	Copyright Hunter Moseley, 1999. All rights reserved.
#	Written by Hunter Moseley 4/2/1999
#
#       Modified Gurmukh Sahota 06/17/2001
#            Allow use of - in new_workbook_name to signify to keep the workbook name from
#            the original peaklist.
#
#
#   usage: extract_duplicates.pl in_peak_file inclusion_toggle new_workbook_name \
# 			       out_peak_file to11 tol2 ...
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 	inclusion_toggle:
# 		inc(lusive) - include only peaks with duplicates.
# 		exc(lusive) - exclude peaks with duplicates.
# 		low 	    - include only the low intensity representative of each set 
# 			      of duplicates.
# 		high 	    - include only the high intensity representative of each set 
# 			      of duplicates.
# 	workbook_name	- string to use as workbook name or "-" to use old name.
# 	tol1 tol2 ... 	- dimension tolerances for comparison or "0" not to use specific 
# 			  dimension.
#
use strict;

# Print usage info
if ((@ARGV < 5) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "extract_duplicates.pl \n";
  print "	Extract a group of peaks that have duplicate chemical shifts.\n";
  print "\n";
  print "  usage: extract_duplicates.pl in_peak_file inclusion_toggle new_workbook_name \\\n";
  print "			       out_peak_file to11 tol2 ...\n";
  print "\n";
  print "	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print "	inclusion_toggle:\n";
  print "		inc(lusive) - include only peaks with duplicates.\n";
  print "		exc(lusive) - exclude peaks with duplicates.\n";
  print "		low 	    - include only the low intensity representative of each set \n";
  print "			      of duplicates.\n";
  print "		high 	    - include only the high intensity representative of each set \n";
  print "			      of duplicates.\n";
  print "	workbook_name	- string to use as workbook name  or \"-\" to use old name.\n";
  print "	tol1 tol2 ... 	- dimension tolerances for comparison or \"0\" not to use specific \n";
  print "			  dimension.\n";
  exit(1);
  }

# Read in comand line parameters
my $in_peak_file = $ARGV[0];
my $inclusion;
if ($ARGV[1] =~ /^inc/)
  { $inclusion = 1; }
elsif ($ARGV[1] =~ /^exc/)
  { $inclusion = 0; }
elsif ($ARGV[1] =~ /^low/)
  { $inclusion = -1; }
else
  { $inclusion = -2; }
my $workbook_name = $ARGV[2];
my $out_peak_file = $ARGV[3];

my $x;
my @tolerance_alist;
for($x = 4; $x < @ARGV; $x++)
  { push @tolerance_alist, $ARGV[$x]; }

# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);
my $num_cols = @{$$peak_alist[0]};

# Create Equivalence array
my @equiv_alist;
my $equiv_count = 1;
my $pindex1;
my $duplicate_alist;
for($pindex1 = 0; $pindex1 < @$peak_alist; $pindex1++)
  {
  my $duplicate_alist = &find_duplicate_index_list($pindex1, $peak_alist, \@tolerance_alist);
  if (@$duplicate_alist)
    {
    if (! $equiv_alist[$pindex1])
      { $equiv_alist[$pindex1] = $equiv_count++; }

    my $dup_index;
    foreach $dup_index (@$duplicate_alist)
      {
      if (! $equiv_alist[$dup_index])
	{ $equiv_alist[$dup_index] = $equiv_alist[$pindex1]; }
      }
    }
  }


# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x < $num_cols-2; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print duplicate peak list.
my $pindex;
my $x = 1;
for($pindex = 0; $pindex < @$peak_alist; $pindex++)
  {
  if ((! $equiv_alist[$pindex]) == (! $inclusion)) # is the peak to be printed?
    {
    my $equiv_value = $equiv_alist[$pindex];
    my $pindex2;
    my $best_index = $pindex; # for use for high/low inclusion.
    for($pindex2 = $pindex; $pindex2 < @$peak_alist; $pindex2++)
      {
      if ($equiv_alist[$pindex2] == $equiv_value)
	{
	$equiv_alist[$pindex2] = ! $inclusion; # set the peak as being no longer useful.

	if ($inclusion >= 0) # print peaks if inclusion is inc/exc
	{
	    $$peak_alist[$pindex2][0] = $x++; 
	    $$peak_alist[$pindex2][$num_cols-2] = "\t" . $$peak_alist[$pindex2][$num_cols-2];

	    # GSS 06/17/2001   added if statement to allow for the original workbook names to be kept.
	    if ($workbook_name eq "-")
	    {	
		$$peak_alist[$pindex2][$num_cols-1] = ("\t" x (length($$peak_alist[$pindex2][$num_cols-2]) < 9)) . $$peak_alist[$pindex2][$num_cols-1]; 
	    }
	    else
	    { 
		my ($notes) = ($$peak_alist[$pindex2][$num_cols-1] =~ /[^.]*[.](.*)/);
		$$peak_alist[$pindex2][$num_cols-1] = ("\t" x (length($$peak_alist[$pindex2][$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes; 
	    }
	  print OUTPUT_FILE join("\t",@{$$peak_alist[$pindex2]}), "\n";
	  }
	elsif (($inclusion == -1) && ($$peak_alist[$pindex2][$num_cols-2] < $$peak_alist[$best_index][$num_cols-2])) # find lowest intensity value
	  { $best_index = $pindex2; }
	elsif (($inclusion == -2) && ($$peak_alist[$pindex2][$num_cols-2] > $$peak_alist[$best_index][$num_cols-2])) # find highest intensity value
	  { $best_index = $pindex2; }
	}
      }

    if ($inclusion == 1) # print duplicate set divider
      { print OUTPUT_FILE "#\n"; }
    elsif ($inclusion < 0) # print high/low peak
      { 
      $$peak_alist[$best_index][0] = $x++; 
      $$peak_alist[$best_index][$num_cols-2] = "\t" . $$peak_alist[$best_index][$num_cols-2];

      # GSS 06/17/2001   added if statement to allow for the original workbook names to be kept.
      if ($workbook_name eq "-")
      {	
	  $$peak_alist[$best_index][$num_cols-1] = ("\t" x (length($$peak_alist[$best_index][$num_cols-2]) < 9)) . $$peak_alist[$best_index][$num_cols-1]; 
      }
      else
      { 
      my ($notes) = ($$peak_alist[$best_index][$num_cols-1] =~ /[^.]*[.](.*)/);
      $$peak_alist[$best_index][$num_cols-1] = ("\t" x (length($$peak_alist[$best_index][$num_cols-2]) < 9)) . $workbook_name . ("."x($notes ne "")) . $notes; 
      }
      print OUTPUT_FILE join("\t",@{$$peak_alist[$best_index]}), "\n";
      }
    }
  }

# Print Comments
print OUTPUT_FILE "*\n";

my $line;
my $history_found = 0;
foreach $line (@$end_comments_alist)
  {
  if (! $history_found && ($line =~ /^File History:/))
    {
    print OUTPUT_FILE "File History: $out_peak_file \n\n";
    print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
    $history_found = 1;
    next;
    }
  
  print OUTPUT_FILE $line;  
  }

if (! $history_found)
  {
  print OUTPUT_FILE "\nFile History: $out_peak_file \n\n";
  print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
  }

close OUTPUT_FILE;


#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

sub find_duplicate_index_list
  {
  my $pindex = shift @_;
  my $peak_alist = shift @_;
  my $tolerance_alist = shift @_;

  my $duplicate_alist = [];
  my $x;
  my $y;
  for($y = $pindex+1; $y < @$peak_alist; $y++)
    {
    for($x=0; $x < @$tolerance_alist; $x++)
      {
      if (($$tolerance_alist[$x] != 0.0) && (abs($$peak_alist[$pindex][$x+1] - $$peak_alist[$y][$x+1]) >= $$tolerance_alist[$x]))
	{ last; }
      }

    if ($x == @$tolerance_alist)
      { push @$duplicate_alist, $y; }
    }

  return $duplicate_alist;
  }
