#!/usr/bin/perl
#!/usr/bin/perl
# file_history.pl 
# 	Print File History of a peak file.
# 
#	Copyright Hunter Moseley, 2000. All rights reserved.
#	Written by Hunter Moseley 9/20/2000
#
#   usage: file_history.pl in_peak_file
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
#
use strict;

# Print usage info
if ((@ARGV < 1) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "file_history.pl \n";
  print " 	Print File History of a peak file.\n";
  print " \n";
  print "  usage: file_history.pl in_peak_file\n";
  print "\n";
  print " 	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  exit(1);
  }

# Read in comand line parameters
my $in_peak_file = $ARGV[0];


# Read Peak files
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);

# Skip excess comments
while(@$end_comments_alist && ($$end_comments_alist[0] !~ /File History/))
  { shift @$end_comments_alist; }

# Print File History if present
print "\n", @$end_comments_alist;


#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

