#!/usr/bin/perl
#!/usr/bin/perl
# jval2bmrb.pl
# 	Converts AutoAssign peak list into BMRB spectral peak list save frame.
# 
#	Copyright Gurmukh Sahota, 2002. All rights reserved.
#	Written by Gurmukh Sahota 5/13/2002
#
#   usage: jval2bmrb.pl in_j_file in_bmrb_file out_bmrb_file
# 
# 	j_file 	      - name of j file or "-" for standard input/output.
# 	bmrb_file     - name of BMRB file.
#
#
#  Updated to use new BMRBParsing/CMapParsing modules.
#  Copyright Hunter Moseley 2005. All rights reserved.
#
#  Hunter Moseley: Updated program to not overwrite an existing file.
#

#
# These modules allow me to use modules rather than hardcoding everything into one script
use FindBin;
use lib $FindBin::Bin;

# BMRB Parser
use BMRBParsing qw(:ALL);

use strict;

#
# Print usage info
if ((@ARGV < 3) || ($ARGV[0] =~ /^-h/))
  {
  print STDERR "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print STDERR "\n";
  print STDERR " jval2bmrb.pl\n";
  print STDERR " 	Converts AutoStructure Input J HNHA values into BMRB coupling constants save frame.\n";
  print STDERR " \n";
  print STDERR "   usage: jval2bmrb.pl in_j_file in_bmrb_file out_bmrb_file\n";
  print STDERR " \n";
  print STDERR " 	j_file 	         - name of j file or \"-\" for standard input/output.\n";
  print STDERR " 	bmrb_file 	 - name of BMRB file.\n";
  print STDERR "\n";
  exit(1);
  }

# Read command line arguements.
my $in_j_file = shift @ARGV;
my $in_bmrb_file = shift @ARGV;
my $out_bmrb_file = shift @ARGV;

#while (($ARGV[0] =~ /^\-/) && @ARGV)
#  {
#  my $switch = shift @ARGV;
#  if ($switch =~ /^\-a/)
#    { $append_to_bmrb = 1; }
#  else
#    { 
#    print STDERR "INVALID OPTION: $switch\n\nExiting Program ...\n\n";  
#    exit(1);
#    }
#  }

my $bmrb_hlist = &read_bmrb_file($in_bmrb_file); 

local *JFILE;
if ($in_j_file =~ /^\-$/)
  { *JFILE = *STDIN; }
else
  { open(JFILE, "<$in_j_file") || die("Unable to open input file: " . $in_j_file) ; }


my $column_elements = {};

# Names based on an older NMR-Star 3.0 version.  Do not use in BMRBParsing
# otherwise they are good for what their current use is, as descriptive
# variable names

$$column_elements{"_Coupling_constant_code"}        = 1;
$$column_elements{"_Atom_one_residue_seq_code"}     = 2;
$$column_elements{"_Atom_one_residue_label"}        = 3;
$$column_elements{"_Atom_one_atom_name"}            = 4;
$$column_elements{"_Atom_two_residue_seq_code"}     = 5;
$$column_elements{"_Atom_two_residue_label"}        = 6;
$$column_elements{"_Atom_two_atom_name"}            = 7;
$$column_elements{"_Coupling_constant_value"}       = 8;
$$column_elements{"_Coupling_constant_value_error"} = 9;

my $last_id = 0;
foreach my $test_list (values %{$$bmrb_hlist{"cc_lists"}})
  {
  if ($$test_list{"cc_list_id"} > $last_id)
    { $last_id = $$test_list{"cc_list_id"}; }
  }

my $cc_list = { "elements" => [], "cc_list_id" => $last_id+1 };

my $count= 0;
while (my $line = <JFILE>)
  {
  next   if ($line =~ /^\s*$/);
  my @tokens = split /\s+/, $line;
  
  my $res_name1 = &convert_aa3to1($tokens[$$column_elements{"_Atom_one_residue_label"}]) . $tokens[$$column_elements{"_Atom_one_residue_seq_code"}];
  my $entity1 = &find_entity($bmrb_hlist, $res_name1);
  my $res_name2 = &convert_aa3to1($tokens[$$column_elements{"_Atom_two_residue_label"}]) . $tokens[$$column_elements{"_Atom_two_residue_seq_code"}];
  my $entity2 = &find_entity($bmrb_hlist, $res_name1);
  next if (! $entity1 || ! $entity2);

  $$cc_list{"elements"}[$count]{"id"}                 = $count + 1;
  $$cc_list{"elements"}[$count]{"type"}               = $tokens[$$column_elements{"_Coupling_constant_code"}];
  $$cc_list{"elements"}[$count]{"Atom1entityID"}      = $$entity1{"entity_id"};
  $$cc_list{"elements"}[$count]{"Atom1residueNumber"} = $tokens[$$column_elements{"_Atom_one_residue_seq_code"}];
  $$cc_list{"elements"}[$count]{"Atom1residueName"}   = &convert_aa3to1($tokens[$$column_elements{"_Atom_one_residue_label"}]);
  $$cc_list{"elements"}[$count]{"Atom1atomName"}      = $tokens[$$column_elements{"_Atom_one_atom_name"}];
  $$cc_list{"elements"}[$count]{"Atom1atomType"}      = substr($tokens[$$column_elements{"_Atom_one_atom_name"}],0,1);
  $$cc_list{"elements"}[$count]{"Atom2entityID"}      = $$entity2{"entity_id"};
  $$cc_list{"elements"}[$count]{"Atom2residueNumber"} = $tokens[$$column_elements{"_Atom_two_residue_seq_code"}];
  $$cc_list{"elements"}[$count]{"Atom2residueName"}   = &convert_aa3to1($tokens[$$column_elements{"_Atom_two_residue_label"}]);
  $$cc_list{"elements"}[$count]{"Atom2atomName"}      = $tokens[$$column_elements{"_Atom_two_atom_name"}];
  $$cc_list{"elements"}[$count]{"Atom2atomType"}      = substr($tokens[$$column_elements{"_Atom_two_atom_name"}],0,1);
  $$cc_list{"elements"}[$count]{"value"}              = $tokens[$$column_elements{"_Coupling_constant_value"}];
  $$cc_list{"elements"}[$count]{"error"}              = $tokens[$$column_elements{"_Coupling_constant_value_error"}];
  $$cc_list{"elements"}[$count]{"Atom1ambiguity"}      = "1";
  $$cc_list{"elements"}[$count]{"Atom2ambiguity"}      = "1";
  
  if (&convert_aa3to1($tokens[$$column_elements{"_Atom_one_residue_label"}]) eq "G")
    {
    if ($$cc_list{"elements"}[$count]{"Atom1atomName"} eq "H")
      {
      $$cc_list{"elements"}[$count]{"Atom2atomName"}      = "HA2";
      $$cc_list{"elements"}[$count]{"Atom2ambiguity"}      = "2";
      }
    elsif ($$cc_list{"elements"}[$count]{"Atom2atomName"} eq "H")
      {
      $$cc_list{"elements"}[$count]{"Atom1atomName"}      = "HA2";
      $$cc_list{"elements"}[$count]{"Atom1ambiguity"}      = "2";
      }
    
    $count++;
    
    $$cc_list{"elements"}[$count]{"id"}                 = $count + 1;
    $$cc_list{"elements"}[$count]{"type"}               = $tokens[$$column_elements{"_Coupling_constant_code"}];
    $$cc_list{"elements"}[$count]{"Atom1entityID"}      = $$entity1{"entity_id"};
    $$cc_list{"elements"}[$count]{"Atom1residueNumber"} = $tokens[$$column_elements{"_Atom_one_residue_seq_code"}];
    $$cc_list{"elements"}[$count]{"Atom1residueName"}   = &convert_aa3to1($tokens[$$column_elements{"_Atom_one_residue_label"}]);
    $$cc_list{"elements"}[$count]{"Atom1atomName"}      = $tokens[$$column_elements{"_Atom_one_atom_name"}];
    $$cc_list{"elements"}[$count]{"Atom1atomType"}      = substr($tokens[$$column_elements{"_Atom_one_atom_name"}],0,1);
    $$cc_list{"elements"}[$count]{"Atom2entityID"}      = $$entity2{"entity_id"};
    $$cc_list{"elements"}[$count]{"Atom2residueNumber"} = $tokens[$$column_elements{"_Atom_two_residue_seq_code"}];
    $$cc_list{"elements"}[$count]{"Atom2residueName"}   = &convert_aa3to1($tokens[$$column_elements{"_Atom_two_residue_label"}]);
    $$cc_list{"elements"}[$count]{"Atom2atomName"}      = $tokens[$$column_elements{"_Atom_two_atom_name"}];
    $$cc_list{"elements"}[$count]{"Atom2atomType"}      = substr($tokens[$$column_elements{"_Atom_two_atom_name"}],0,1);
    $$cc_list{"elements"}[$count]{"value"}              = $tokens[$$column_elements{"_Coupling_constant_value"}];
    $$cc_list{"elements"}[$count]{"error"}              = $tokens[$$column_elements{"_Coupling_constant_value_error"}];
    $$cc_list{"elements"}[$count]{"Atom1ambiguity"}      = "1";
    $$cc_list{"elements"}[$count]{"Atom2ambiguity"}      = "1";
    
    if ($$cc_list{"elements"}[$count]{"Atom1atomName"} eq "H")
      {
      $$cc_list{"elements"}[$count]{"Atom2atomName"}      = "HA3";
      $$cc_list{"elements"}[$count]{"Atom2ambiguity"}      = "2";
      }
    elsif ($$cc_list{"elements"}[$count]{"Atom2atomName"} eq "H")
      {
      $$cc_list{"elements"}[$count]{"Atom1atomName"}      = "HA3";
      $$cc_list{"elements"}[$count]{"Atom1ambiguity"}      = "2";
      }
    }
  
  $count++;
  }

close JFILE;

if ($count > 0)
  { $$bmrb_hlist{"cc_lists"}{$$cc_list{"cc_list_id"}} = $cc_list; }

#
# write bmrb file out
#
&write_bmrb_file($bmrb_hlist, $out_bmrb_file);
