#!/usr/bin/perl
#!/usr/bin/perl
#  manage_bmrb.pl
#	Manage the save frames in a bmrb file.
#
#  usage: manage_bmrb.pl input_bmrb_file [options]
#
#		input_bmrb_file - bmrb file to read ("-" to read from standard input).
#		options:
#			-add bmrb_file save_frame_list - add list of comma-seperated save frames from the bmrb_file.
#			-delete save_frame_list - delete the list of comma-separated save frames.
#			-extract save_frame_alist - extract comma-separated list of save frames in the given order.
#			-output bmrb_file - bmrb_file to save changes to ("-" to write to standard output).
#
#
#  Written by Hunter Moseley 12/3/2001
#  Copyright Hunter Moseley, 2001. All rights reserved.
#
#
#  Updated to use new BMRBParsing/CMapParsing modules.
#  Copyright Hunter Moseley 2005. All rights reserved.
#

# These modules allow me to use modules rather than hardcoding everything into one script
use FindBin;
use lib $FindBin::Bin;

# Parser module parses in/out the cmap data and the peak file
use BMRBParsing qw(:ALL);

# Use the strict module: This is off because of silly bareword problems with the initialize_matirices
 use strict;

if ((@ARGV < 1) || ($ARGV[0] =~ /^-h/))
  {
  print STDERR "Error: not enough parameters given.\n\n" if (@ARGV < 1);
  print STDERR "manage_bmrb.pl\n";
  print STDERR "	Manage the save frames in a bmrb file.\n";
  print STDERR "\n";
  print STDERR "  usage: manage_bmrb.pl input_bmrb_file [options]\n";
  print STDERR "\n";
  print STDERR "		input_bmrb_file - bmrb file to read (\"-\" to read from standard input).\n";
  print STDERR "		options:\n";
  print STDERR "			-add bmrb_file save_frame_list - add list of comma-seperated save frames from the bmrb_file.\n";
  print STDERR "			-delete save_frame_list - delete the list of comma-separated save frames.\n";
  print STDERR "			-extract save_frame_alist - extract comma-separated list of save frames in the given order.\n";
  print STDERR "			-output bmrb_file - bmrb_file to save changes to (\"-\" to write to standard output).\n";
  print STDERR "\n";
  exit(1);
  }

# read command line options
my $infile = shift @ARGV;
my $add_filename = "";
my $add_alist;
my $delete_alist;
my $output_filename = "";
my $extract_alist;

for(my $x=0; $x < @ARGV; $x++) 
  {
  if ($ARGV[$x] =~ /-a/)
    { 
    $x++;
    $add_filename = $ARGV[$x];
    $x++;
    $add_alist = [ split(/[,]/,$ARGV[$x]) ];
    }
  elsif ($ARGV[$x] =~ /^-d/)
    { $x++; $delete_alist = [ sort { $b <=> $a; } (split(/[,]/,$ARGV[$x])) ]; }
  elsif ($ARGV[$x] =~ /^-o/)
    { $x++; $output_filename = $ARGV[$x]; }
  elsif ($ARGV[$x] =~ /^-e/)
    { $x++; $extract_alist = [ split(/[,]/,$ARGV[$x]) ]; }
  else
    {
    print STDERR "Error: unknown option ",$ARGV[$x]," given.\n";
    exit(1);
    }
  }

if (( (ref $delete_alist eq "ARRAY") || (ref $add_alist eq "ARRAY") || (ref $add_alist eq "ARRAY") || (ref $extract_alist eq "ARRAY")) && ($output_filename eq ""))
  { print STDERR "\nWarning: no output file given with modifying options.\n\n"; }



# read bmrb file
my $bmrb_hlist = &read_bmrb_file($infile, { "skip_processing" => 1 });

# delete save frames if indicated.
if (ref $delete_alist eq "ARRAY")
  {
  foreach my $delete_frame (@$delete_alist)
    {
    if (($delete_frame <= @{$$bmrb_hlist{"save_frames"}}) && ($delete_frame > 0))
      { splice(@{$$bmrb_hlist{"save_frames"}},($delete_frame-1),1); }
    }
  }

# add save frames if indicated.
if (ref $add_alist eq "ARRAY")
  {
  my $bmrb_hlist2 = &read_bmrb_file($add_filename, { "skip_processing" => 1 });

  foreach my $add_frame (@$add_alist)
    {
    if (($add_frame <= @{$$bmrb_hlist2{"save_frames"}}) && ($add_frame > 0))
      { push @{$$bmrb_hlist{"save_frames"}}, $$bmrb_hlist2{"save_frames"}[$add_frame-1]; }
    }
  }


# extract save frames
if (ref $extract_alist eq "ARRAY")
  {
  my $temp_alist = [];

  foreach my $extract_frame (@$extract_alist)
    {
    if (($extract_frame <= @{$$bmrb_hlist{"save_frames"}}) && ($extract_frame > 0))
      { push @$temp_alist, $$bmrb_hlist{"save_frames"}[$extract_frame-1]; }
    }  

  $$bmrb_hlist{"save_frames"} = $temp_alist;
  }

# print a list of save frames
print "\nList of save frames in $infile\n----------------------------------\n";
for(my $count=1; $count <= @{$$bmrb_hlist{"save_frames"}}; $count++)
  {  printf "%3d) %s - %s\n",$count,$$bmrb_hlist{"save_frames"}[$count-1]{"type"},$$bmrb_hlist{"save_frames"}[$count-1]{"code"};  }  

if ($output_filename ne "")
  { &write_bmrb_file($bmrb_hlist, $output_filename); }

