#!/usr/bin/perl
#!/usr/bin/perl
#
# pales2cns.pl
#	Converts data from pales output table format to cns input format for RDCs.
#
#	Usage: pales2cns.pl input_file output_file extract_toggle [option]
#		input_file		- PALES output table of measured and predicted RDCs.
#		output_file		- CNS formatted file for RDCs.
#		extract_toggle:
#			meas(ured)	- measured RDCs.
#			pred(icted)	- RDCs predicted by PALES.
#		options:
#			-margin n - n is the error margin to use in CNS calculations.
#
# IMPORTANT: Deyou's original script took out predicted RDCs, while this one takes out either.
# 
# Copyright Aneerban Bhattacharya, 2001. All rights reserved.
# Originally by Deyou Zheng
# Modified by Aneerban Bhattacharya, Apr 4, 2001
# Modified by Aneerban Bhattacharya	Mar 22, 2002
# 
# Interactive help format taken from AutoPeak scripts written by Hunter Moseley.
#
#

use strict;


if ((scalar(@ARGV) < 3) || ($ARGV[0] =~ /^-h/)) {
    print(STDERR "\nError: not enough parameters given.\n") if ($ARGV[0] !~ /^-h/);
    print(STDERR "\n");
	print(STDERR "pales2cns.pl\n");
	print(STDERR "\tConverts data from pales output table format to cns input format for RDCs.\n");
	print(STDERR "\n");
	print(STDERR "\tUsage: pales2cns.pl input_file output_file extract_toggle [option]\n");
	print(STDERR "\t\tinput_file\t\t- PALES output table of measured and predicted RDCs.\n");
	print(STDERR "\t\toutput_file\t\t- CNS formatted file for RDCs.\n");
	print(STDERR "\t\textract_toggle:\n");
	print(STDERR "\t\t\tmeas(ured)\t- measured RDCs.\n");
	print(STDERR "\t\t\tpred(icted)\t- RDCs predicted by PALES.\n");
	print(STDERR "\t\toptions:\n");
    print(STDERR "\t\t\t-margin n - n is the error margin to use in CNS calculations.\n");
	print(STDERR "\n");
	exit(1);
}


my ($infile_s, $outfile_s, $extract_toggle_b) = @ARGV;

my $line_s;
my $error_margin_n = 0.500; # margin to use in CNS calculations

if ((scalar(@ARGV) > 4) && ($ARGV[3] =~ /^-mar/) && $ARGV[4] =~ /^\-?\d*\.\d+$/) {
    $error_margin_n = $ARGV[4];
}

open (IF, "<$infile_s") || die ("Can't open $infile_s :  $!\n");
open (OF, ">$outfile_s") || die ("Can't open $outfile_s :  $!\n");

while ($line_s = <IF>) {
	if ($line_s =~ /^\s*\d+\s+[A-Z]{3}/) {
		$line_s =~ s/^\s+(.*)/$1/; # remove leading whitespace
    	chomp($line_s);

		my @tokens_a = split(/\s+/, $line_s);
		my $RDC_n = 0.000;
		
		if ($extract_toggle_b =~ /^meas/) { # get measured RDC values
			$RDC_n = $tokens_a[7];
		}
		elsif ($extract_toggle_b =~ /^pred/) { # get predicted RDCs
			$RDC_n = $tokens_a[8];
		}
		
		print(OF "assign ( resid 500  and name OO  )\n       ( resid 500  and name Z   )\n       ( resid 500  and name X   )\n       ( resid 500  and name Y   )\n");
		print(OF sprintf("       ( resid %3d  and name N   )\n", $tokens_a[0]));
		print(OF sprintf("       ( resid %3d  and name HN  )  %.4f %.4f \n\n", $tokens_a[0], $RDC_n, $error_margin_n));
    }
}

close(OF);
close(IF);
