#!/usr/bin/perl
#!/usr/bin/perl
# pks2SparkyPks.pl
# 	Convert an AutoAssign peak list file to a Sparky peak list file.
# 
#   usage: pks2SparkyPks.pl in_peak_file out_peak_file
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 
#	Copyright Hunter Moseley, 2002. All rights reserved.
#	Written by Hunter Moseley 10/7/2002
#
use strict;

# Print usage info
if ((@ARGV < 2) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print " pks2SparkyPks.pl\n";
  print " 	Convert an AutoAssign peak list file to a Sparky peak list file.\n";
  print " \n";
  print "   usage: pks2SparkyPks.pl in_peak_file out_peak_file\n";
  print "\n";
  print " 	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print " \n";
  exit(1);
  }

# Read in command line parameters
my $input_peak_file = $ARGV[0];
my $out_peak_file = $ARGV[1];

# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($input_peak_file);
my $num_columns = @{$$peak_alist[0]};

# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }


# Print peaks in Sparky format
my $blank_sparky_assignment = "?" . ("-?" x ($num_columns - 4));
foreach my $peak (@$peak_alist)
  { 
  print OUTPUT_FILE $blank_sparky_assignment,"\t ";
  for(my $x=1; $x < $num_columns - 2; $x++)
    { print OUTPUT_FILE $$peak[$x], "\t "; }
  print OUTPUT_FILE $$peak[$num_columns-2],"\n";
  }

close OUTPUT_FILE;

#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

