#!/usr/bin/perl
#!/usr/bin/perl
# remove_peaks.pl 
# 	Removes peaks from a peak list with the given indeces.
# 
#	Copyright Hunter Moseley, 2002. All rights reserved.
#	Written by Hunter Moseley 1/13/2002
#
#   usage: remove_peaks.pl in_peak_file out_peak_file index1 index2 ...
# 
# 	peak_file	  - name of peak file or "-" for standard input/output.
# 	index1 index2 ... - list of peak indeces to remove.
#
use strict;

# Print usage info
if ((@ARGV < 3) || ($ARGV[0] =~ /^-h/))
  {
  print STDERR "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print STDERR "\n";
  print STDERR " remove_peaks.pl \n";
  print STDERR " 	Removes peaks from a peak list with the given indeces.\n";
  print STDERR " \n";
  print STDERR "   usage: remove_peaks.pl in_peak_file out_peak_file index1 index2 ...\n";
  print STDERR " \n";
  print STDERR " 	peak_file	  - name of peak file or \"-\" for standard input/output.\n";
  print STDERR " 	index1 index2 ... - list of peak indeces to remove.\n";
  exit(1);
  }

# Read in comand line parameters
my $in_peak_file = $ARGV[0];
my $out_peak_file = $ARGV[1];

my $x;
my @remove_alist;
for($x = 2; $x < @ARGV; $x++)
  { push @remove_alist, $ARGV[$x]; }


# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);
my $num_cols = @{$$peak_alist[0]};

# Open Output file
local *OUTPUT_FILE;
if ($out_peak_file eq "-")
  { *OUTPUT_FILE = *STDOUT; }
else
  { open(OUTPUT_FILE, ">$out_peak_file") || die ("open " . $out_peak_file . " failed."); }

# Print Header
print OUTPUT_FILE "#Index	";
my $x;
for ($x=1; $x < $num_cols-2; $x++)
  { print OUTPUT_FILE $x, "Dim	"; }
print OUTPUT_FILE "	Intensity	Workbook\n";

# Print filtered peak list. 
my $peak;
foreach $peak (@$peak_alist) 
  {
  $$peak[$num_cols-2] = "\t" . $$peak[$num_cols-2];
  $$peak[$num_cols-1] = ("\t" x (length($$peak[$num_cols-2]) < 9)) . $$peak[$num_cols-1];
  if (grep { $_ == $$peak[0]; } (@remove_alist))
    { print OUTPUT_FILE "#"; }
  
  print OUTPUT_FILE join("\t",@$peak), "\n";
  }

# Print Comments
print OUTPUT_FILE "*\n";

my $line;
my $history_found = 0;
foreach $line (@$end_comments_alist)
  {
  if (! $history_found && ($line =~ /^File History:/))
    {
    print OUTPUT_FILE "File History: $out_peak_file \n\n";
    print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
    $history_found = 1;
    next;
    }
  
  print OUTPUT_FILE $line;  
  }

if (! $history_found)
  {
  print OUTPUT_FILE "\nFile History: $out_peak_file \n\n";
  print OUTPUT_FILE $0, " ", join(' ',@ARGV),"\n";
  }

close OUTPUT_FILE;


#
#  Subroutines
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }

