#!/usr/bin/perl
#!/usr/bin/perl
# sec2bmrb.pl
#       Converts AutoStructure Output .sec file into BMRB secondary_structure save frame.
# 
#	Copyright Gurmukh Sahota, 2002. All rights reserved.
#	Written by Gurmukh Sahota 5/13/2002
#
#   usage: sec2bmrb.pl in_sec_file in_bmrb_file out_bmrb_file
# 
# 	sec_file      - name of sec file or "-" for standard input/output.
# 	bmrb_file     - name of BMRB file.
#
#
#  Updated to use new BMRBParsing/CMapParsing modules.
#  Copyright Hunter Moseley 2005. All rights reserved.
#
#  
#  Hunter Moseley: Updated program to not overwrite an existing file.
#  Had to rewrite most of the program on 08/22/2005
#

# allows perl to fine local modules.
use FindBin;
use lib $FindBin::Bin;

# contains subroutines for reading and writing bmrb files.
use BMRBParsing qw(:ALL);

my %BMRBConvertionHash = (Coil  => "loop       ",
			  Helix => "alpha-helix",
			  Sheet => "beta-sheet ");

# Print usage info
if ((@ARGV < 2) || ($ARGV[0] =~ /^-h/))
  {
  print STDERR "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print STDERR "\n";
  print STDERR " sec2bmrb.pl\n";
  print STDERR " 	Converts AutoStructure output .sec file into BMRB secondary_structure save frame.\n";
  print STDERR " \n";
  print STDERR "   usage: sec2bmrb.pl in_sec_file in_bmrb_file out_bmrb_file\n";
  print STDERR " \n";
  print STDERR " 	sec_file 	 - name of sec file or \"-\" for standard input/output.\n";
  print STDERR " 	bmrb_file 	 - name of BMRB file.\n";
  print STDERR "\n";
  exit(1);
  }

# Read command line arguements.
my $in_sec_file = shift @ARGV;
my $in_bmrb_file = shift @ARGV;
my $out_bmrb_file = shift @ARGV;

while (($ARGV[0] =~ /^\-/) && @ARGV)
  {
  my $switch = shift @ARGV;
  if (1)
    { 
    print STDERR "INVALID OPTION: $switch\n\nExiting Program ...\n\n";  
    exit(1);
    }
  }

# read bmrb file
my $bmrb_hlist = &read_bmrb_file($in_bmrb_file); 

# determine save secondary structure id
# should not be executed, but to keep uniformity and increase # lines of code
my $secondary_structure_id = 0;
foreach my $test_id (keys %{$$bmrb_hlist{"ss_lists"}})
  { 
  if ($test_id > $secondary_structure_id)
    { $secondary_structure_id = $test_id; }
  }
$secondary_structure_id++;

local *SEC_FILE;
if ($in_sec_file =~ /^\-$/)
  { *SEC_FILE = *STDIN; }
else
  { open(SEC_FILE, "<$in_sec_file") || die("Unable to open input file: " . $in_sec_file); }

# Read file and get residue info if necessary
my %idAA;
my %SSE;
while (my $line = <SEC_FILE>)
  {
  chomp $line;
  if ($line =~ /^Summary:/)
    {
    <SEC_FILE>;  <SEC_FILE>;
    while ($line = <SEC_FILE>)
      {
      chomp $line;
      last if ($line =~ /^\s*$/);
      my @elem = split /\s+/, $line;
      $idAA{$elem[0]} = $elem[1] . $elem[0];
      }
    }

  if ($line =~ /^Final Secondary/)
    {
    <SEC_FILE>; <SEC_FILE>;
    while ($line = <SEC_FILE>)
      {
      chomp $line;
      last if ($line =~ /^\s*$/);
      my @elem = split /\s+/, $line;
      $elem[0] =~ /^ID=(\d+)$/;
      $SSE{$1} = [$elem[1], $elem[2]];
      }       
    }
  }

close SEC_FILE;

# setup the SSE into BMRB format
my $ss_list = { "ss_list_id" => $secondary_structure_id, "elements" => [] };
foreach my $id (sort {$a<=>$b} keys %SSE)
  {
  my ($start_index, $end_index) = ($SSE{$id}[1] =~ /^(\d+)-(\d+)$/);
  my $entity = &find_entity($bmrb_hlist, $idAA{$start_index});
  
  if ($entity)
    {
    push @{$$ss_list{"elements"}}, 
      { 
      "id" => $id, 
      "start_index" => $start_index, 
      "entity_id" => $$entity{"entity_id"}, 
      "end_index" => $end_index, 
      "sseCode" =>  $BMRBConvertionHash{$SSE{$id}[0]}, 
      "method" => "AutoStructure" 
	};
    }
  }

$$bmrb_hlist{"ss_lists"}{$secondary_structure_id} = $ss_list;

# write bmrb file out
&write_bmrb_file($bmrb_hlist, $out_bmrb_file);
