#!/usr/bin/perl
#!/usr/bin/perl
# sparkyPks2pks.pl 
# 	Create a peak list given the shift values and intensity in sparky format.  
# 	It also tries to assert a HN, N15, HX, CX, CO and X dimension ordering.  
# 
#	Copyright Gurmukh Sahota, 2002. All rights reserved.
#	Written by Gurmukh Sahota 05/20/2002
#
#   usage: sparkyPks2pks.pl in_peak_file new_workbook_name out_peak_file
#
# 	peak_file 	- name of peak file or "-" for standard input/output.
# 	workbook_name	- string to use as workbook name.
# 
use strict;

# Print usage info
if ((@ARGV < 3) || ($ARGV[0] =~ /^-h/))
  {
  print "\nError: not enough parameters given.\n" if ($ARGV[0] !~ /^-h/);
  print "\n";
  print "sparkyPks2pks.pl \n";
  print " 	Create a peak list given the shift values and intensity in separate columns.\n";
  print "	It also tries to assert a HN, N15, HX, CX, CO and X dimension ordering.\n";
  print "\n";
  print "  usage: sparkyPks2pks.pl in_peak_file new_workbook_name out_peak_file\n";
  print "\n";
  print "	peak_file 	- name of peak file or \"-\" for standard input/output.\n";
  print "	workbook_name	- string to use as workbook name.\n";
  exit(1);
  }


use FindBin;                    
use lib $FindBin::Bin;    

use PeakParsing qw(:ALL);
use strict;


my %dimorder = (HN=>1, N15=>2, HX=>3, CX=>4, CO=>5, X=>6);

my %ranges = (HN=>[6,10],
	      N15=>[105,135],
	      HX=>[1,5.5],
	      CX=>[20,60],
	      CO=>[160,190]);


my $commandargs = join (" ", @ARGV);

my $sparky_infile = shift @ARGV;
my $workbook_name = shift @ARGV;
my $peak_outfile = shift @ARGV;

my $sparky_peakfile_hlist = &parseSparkyPeakfile($sparky_infile);
my @dims = @{$$sparky_peakfile_hlist{"dimlist"}};

my %average_per_dim;
foreach my $peak_index (keys %{$$sparky_peakfile_hlist{"plist"}})
{
    foreach my $dim (@dims)
    {
	$average_per_dim{$dim}+= $$sparky_peakfile_hlist{"plist"}{$peak_index}{"shifts"}{$dim};
    }
    
    $$sparky_peakfile_hlist{"plist"}{$peak_index}{"workbookname"} = $workbook_name;
    if (exists $$sparky_peakfile_hlist{"plist"}{$peak_index}{"sparky_note"})
      { $$sparky_peakfile_hlist{"plist"}{$peak_index}{"workbookname"} .= "." . $$sparky_peakfile_hlist{"plist"}{$peak_index}{"sparky_note"}; }
}
foreach my $dim (@dims)
{
    $average_per_dim{$dim}/= scalar(keys %{$$sparky_peakfile_hlist{"plist"}});
}

my %assign;
foreach my $dim (@dims)
{
    foreach my $atomType (keys %ranges)
    {
	if (($average_per_dim{$dim} > $ranges{$atomType}[0]) && ($average_per_dim{$dim} < $ranges{$atomType}[1]))
	{
	    $assign{$dim} = $atomType;
	    last;
	}
    }
    $assign{$dim} = "X"  if (! (exists $assign{$dim}) );
}

@{$$sparky_peakfile_hlist{"dimlist"}} = sort {$dimorder{$assign{$a}} <=> $dimorder{$assign{$b}}} @dims;

$$sparky_peakfile_hlist{"end_comments"} = "\nFile History: $peak_outfile\n\n$0 $commandargs\n";
&outputPeakfile($sparky_peakfile_hlist, $peak_outfile);

